from random import randint


class Block:
	"""
	A block class. Contains the block type and properties.
	"""

	def __init__(self, name, properties):
		"""
		:param block_type: the namespaced block type.
		:param properties: a dictionary of properties for the block.
		"""

		self.name = name
		self.properties = properties
	
	
	def get_block(self):
		"""
		Gets the block's name and properties.
		"""

		return {"name": self.name, "properties": self.properties}


class BlockPalette:
	"""
	Allows a random choice of two or more blocks.
	"""

	def __init__(self, block_array):
		"""
		Creates the block palette. Takes an array of possible blocks as an
		argument.

		:param block_array: the list of possible blocks
		"""

		if len(block_array) < 1:
			return ValueError("Must have at least one block in the list")

		self.block_array = block_array

	
	def get_block(self):
		"""
		Get a random block from the palette.
		"""

		return self.block_array[
			randint(0, len(self.block_array) - 1)].get_block()


class AcaciaButton(Block):
	def __init__(self):
		super().__init__("acacia_button", {})


class AcaciaDoor(Block):
	def __init__(self):
		super().__init__("acacia_door", {})


class AcaciaFence(Block):
	def __init__(self):
		super().__init__("acacia_fence", {})


class AcaciaFenceGate(Block):
	def __init__(self):
		super().__init__("acacia_fence_gate", {})


class AcaciaLeaves(Block):
	def __init__(self):
		super().__init__("acacia_leaves", {})


class AcaciaLog(Block):
	def __init__(self):
		super().__init__("acacia_log", {})


class AcaciaPlanks(Block):
	def __init__(self):
		super().__init__("acacia_planks", {})


class AcaciaPressurePlate(Block):
	def __init__(self):
		super().__init__("acacia_pressure_plate", {})


class AcaciaSign(Block):
	def __init__(self):
		super().__init__("acacia_sign", {})


class AcaciaSlab(Block):
	def __init__(self):
		super().__init__("acacia_slab", {})


class AcaciaStairs(Block):
	def __init__(self):
		super().__init__("acacia_stairs", {})


class AcaciaTrapdoor(Block):
	def __init__(self):
		super().__init__("acacia_trapdoor", {})


class AcaciaWood(Block):
	def __init__(self):
		super().__init__("acacia_wood", {})


class ActivatorRail(Block):
	def __init__(self):
		super().__init__("activator_rail", {})


class Air(Block):
	def __init__(self):
		super().__init__("air", {})


class AncientDebris(Block):
	def __init__(self):
		super().__init__("ancient_debris", {})


class Andesite(Block):
	def __init__(self):
		super().__init__("andesite", {})


class AndesiteSlab(Block):
	def __init__(self):
		super().__init__("andesite_slab", {})


class AndesiteStairs(Block):
	def __init__(self):
		super().__init__("andesite_stairs", {})


class AndesiteWall(Block):
	def __init__(self):
		super().__init__("andesite_wall", {})


class Anvil(Block):
	def __init__(self):
		super().__init__("anvil", {})


class Barrel(Block):
	def __init__(self):
		super().__init__("barrel", {})


class Basalt(Block):
	def __init__(self):
		super().__init__("basalt", {})


class Barrier(Block):
	def __init__(self):
		super().__init__("barrier", {})


class Beacon(Block):
	def __init__(self):
		super().__init__("beacon", {})


class Bedrock(Block):
	def __init__(self):
		super().__init__("bedrock", {})


class BeeNest(Block):
	def __init__(self):
		super().__init__("bee_nest", {})


class Beehive(Block):
	def __init__(self):
		super().__init__("beehive", {})


class Bell(Block):
	def __init__(self):
		super().__init__("bell", {})


class BirchButton(Block):
	def __init__(self):
		super().__init__("birch_button", {})


class BirchDoor(Block):
	def __init__(self):
		super().__init__("birch_door", {})


class BirchFence(Block):
	def __init__(self):
		super().__init__("birch_fence", {})


class BirchFenceGate(Block):
	def __init__(self):
		super().__init__("birch_fence_gate", {})


class BirchLeaves(Block):
	def __init__(self):
		super().__init__("birch_leaves", {})


class BirchLog(Block):
	def __init__(self):
		super().__init__("birch_log", {})


class BirchPlanks(Block):
	def __init__(self):
		super().__init__("birch_planks", {})


class BirchPressurePlate(Block):
	def __init__(self):
		super().__init__("birch_pressure_plate", {})


class BirchSign(Block):
	def __init__(self):
		super().__init__("birch_sign", {})


class BirchSlab(Block):
	def __init__(self):
		super().__init__("birch_slab", {})


class BirchStairs(Block):
	def __init__(self):
		super().__init__("birch_stairs", {})


class BirchTrapdoor(Block):
	def __init__(self):
		super().__init__("birch_trapdoor", {})


class BirchWood(Block):
	def __init__(self):
		super().__init__("birch_wood", {})


class BlackConcrete(Block):
	def __init__(self):
		super().__init__("black_concrete", {})


class BlackConcretePowder(Block):
	def __init__(self):
		super().__init__("black_concrete_powder", {})


class BlackGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("black_glazed_terracotta", {})


class BlackShulkerBox(Block):
	def __init__(self):
		super().__init__("black_shulker_box", {})


class BlackStainedGlass(Block):
	def __init__(self):
		super().__init__("black_stained_glass", {})


class BlackStainedGlassPane(Block):
	def __init__(self):
		super().__init__("black_stained_glass_pane", {})


class BlackTerracotta(Block):
	def __init__(self):
		super().__init__("black_terracotta", {})


class BlackWool(Block):
	def __init__(self):
		super().__init__("black_wool", {})


class Blackstone(Block):
	def __init__(self):
		super().__init__("blackstone", {})


class BlackstoneSlab(Block):
	def __init__(self):
		super().__init__("blackstone_slab", {})


class BlackstoneStairs(Block):
	def __init__(self):
		super().__init__("blackstone_stairs", {})


class BlackstoneWall(Block):
	def __init__(self):
		super().__init__("blackstone_wall", {})


class BlastFurnace(Block):
	def __init__(self):
		super().__init__("blast_furnace", {})


class BlueConcrete(Block):
	def __init__(self):
		super().__init__("blue_concrete", {})


class BlueConcretePowder(Block):
	def __init__(self):
		super().__init__("blue_concrete_powder", {})


class BlueGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("blue_glazed_terracotta", {})


class BlueIce(Block):
	def __init__(self):
		super().__init__("blue_ice", {})


class BlueShulkerBox(Block):
	def __init__(self):
		super().__init__("blue_shulker_box", {})


class BlueStainedGlass(Block):
	def __init__(self):
		super().__init__("blue_stained_glass", {})


class BlueStainedGlassPane(Block):
	def __init__(self):
		super().__init__("blue_stained_glass_pane", {})


class BlueTerracotta(Block):
	def __init__(self):
		super().__init__("blue_terracotta", {})


class BlueWool(Block):
	def __init__(self):
		super().__init__("blue_wool", {})


class BoneBlock(Block):
	def __init__(self):
		super().__init__("bone_block", {})


class Bookshelf(Block):
	def __init__(self):
		super().__init__("bookshelf", {})


class BrainCoralBlock(Block):
	def __init__(self):
		super().__init__("brain_coral_block", {})


class BrickSlab(Block):
	def __init__(self):
		super().__init__("brick_slab", {})


class BrickStairs(Block):
	def __init__(self):
		super().__init__("brick_stairs", {})


class BrickWall(Block):
	def __init__(self):
		super().__init__("brick_wall", {})


class Bricks(Block):
	def __init__(self):
		super().__init__("bricks", {})


class BrownConcrete(Block):
	def __init__(self):
		super().__init__("brown_concrete", {})


class BrownConcretePowder(Block):
	def __init__(self):
		super().__init__("brown_concrete_powder", {})


class BrownGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("brown_glazed_terracotta", {})


class BrownMushroomBlock(Block):
	def __init__(self):
		super().__init__("brown_mushroom_block", {})


class BrownShulkerBox(Block):
	def __init__(self):
		super().__init__("brown_shulker_box", {})


class BrownStainedGlass(Block):
	def __init__(self):
		super().__init__("brown_stained_glass", {})


class BrownStainedGlassPane(Block):
	def __init__(self):
		super().__init__("brown_stained_glass_pane", {})


class BrownTerracotta(Block):
	def __init__(self):
		super().__init__("brown_terracotta", {})


class BrownWool(Block):
	def __init__(self):
		super().__init__("brown_wool", {})


class BubbleCoralBlock(Block):
	def __init__(self):
		super().__init__("bubble_coral_block", {})


class Cactus(Block):
	def __init__(self):
		super().__init__("cactus", {})


class CartographyTable(Block):
	def __init__(self):
		super().__init__("cartography_table", {})


class CarvedPumpkin(Block):
	def __init__(self):
		super().__init__("carved_pumpkin", {})


class Cauldron(Block):
	def __init__(self):
		super().__init__("cauldron", {})


class Chain(Block):
	def __init__(self):
		super().__init__("chain", {})


class ChainCommandBlock(Block):
	def __init__(self):
		super().__init__("chain_command_block", {})


class Chest(Block):
	def __init__(self):
		super().__init__("chest", {})


class ChippedAnvil(Block):
	def __init__(self):
		super().__init__("chipped_anvil", {})


class ChiseledNetherBricks(Block):
	def __init__(self):
		super().__init__("chiseled_nether_bricks", {})


class ChiseledPolishedBlackstone(Block):
	def __init__(self):
		super().__init__("chiseled_polished_blackstone", {})


class ChiseledQuartzBlock(Block):
	def __init__(self):
		super().__init__("chiseled_quartz_block", {})


class ChiseledRedSandstone(Block):
	def __init__(self):
		super().__init__("chiseled_red_sandstone", {})


class ChiseledSandstone(Block):
	def __init__(self):
		super().__init__("chiseled_sandstone", {})


class ChiseledStoneBricks(Block):
	def __init__(self):
		super().__init__("chiseled_stone_bricks", {})


class ChorusFlower(Block):
	def __init__(self):
		super().__init__("chorus_flower", {})


class ChorusPlant(Block):
	def __init__(self):
		super().__init__("chorus_plant", {})


class Clay(Block):
	def __init__(self):
		super().__init__("clay", {})


class CoalBlock(Block):
	def __init__(self):
		super().__init__("coal_block", {})


class CoalOre(Block):
	def __init__(self):
		super().__init__("coal_ore", {})


class CoarseDirt(Block):
	def __init__(self):
		super().__init__("coarse_dirt", {})


class Cobblestone(Block):
	def __init__(self):
		super().__init__("cobblestone", {})


class CobblestoneSlab(Block):
	def __init__(self):
		super().__init__("cobblestone_slab", {})


class CobblestoneStairs(Block):
	def __init__(self):
		super().__init__("cobblestone_stairs", {})


class CobblestoneWall(Block):
	def __init__(self):
		super().__init__("cobblestone_wall", {})


class Cobweb(Block):
	def __init__(self):
		super().__init__("cobweb", {})


class CommandBlock(Block):
	def __init__(self):
		super().__init__("command_block", {})


class Comparator(Block):
	def __init__(self):
		super().__init__("comparator", {})


class Composter(Block):
	def __init__(self):
		super().__init__("composter", {})


class Conduit(Block):
	def __init__(self):
		super().__init__("conduit", {})


class CrackedNetherBricks(Block):
	def __init__(self):
		super().__init__("cracked_nether_bricks", {})


class CrackedPolishedBlackstoneBricks(Block):
	def __init__(self):
		super().__init__("cracked_polished_blackstone_bricks", {})


class CrackedStoneBricks(Block):
	def __init__(self):
		super().__init__("cracked_stone_bricks", {})


class CraftingTable(Block):
	def __init__(self):
		super().__init__("crafting_table", {})


class CrimsonButton(Block):
	def __init__(self):
		super().__init__("crimson_button", {})


class CrimsonDoor(Block):
	def __init__(self):
		super().__init__("crimson_door", {})


class CrimsonFence(Block):
	def __init__(self):
		super().__init__("crimson_fence", {})


class CrimsonFenceGate(Block):
	def __init__(self):
		super().__init__("crimson_fence_gate", {})


class CrimsonHyphae(Block):
	def __init__(self):
		super().__init__("crimson_hyphae", {})


class CrimsonNylium(Block):
	def __init__(self):
		super().__init__("crimson_nylium", {})


class CrimsonPlanks(Block):
	def __init__(self):
		super().__init__("crimson_planks", {})


class CrimsonPressurePlate(Block):
	def __init__(self):
		super().__init__("crimson_pressure_plate", {})


class CrimsonSign(Block):
	def __init__(self):
		super().__init__("crimson_sign", {})


class CrimsonSlab(Block):
	def __init__(self):
		super().__init__("crimson_slab", {})


class CrimsonStairs(Block):
	def __init__(self):
		super().__init__("crimson_stairs", {})


class CrimsonStem(Block):
	def __init__(self):
		super().__init__("crimson_stem", {})


class CrimsonTrapdoor(Block):
	def __init__(self):
		super().__init__("crimson_trapdoor", {})


class CryingObsidian(Block):
	def __init__(self):
		super().__init__("crying_obsidian", {})


class CutRedSandstone(Block):
	def __init__(self):
		super().__init__("cut_red_sandstone", {})


class CutRedSandstoneSlab(Block):
	def __init__(self):
		super().__init__("cut_red_sandstone_slab", {})


class CutSandstone(Block):
	def __init__(self):
		super().__init__("cut_sandstone", {})


class CutSandstoneSlab(Block):
	def __init__(self):
		super().__init__("cut_sandstone_slab", {})


class CyanConcrete(Block):
	def __init__(self):
		super().__init__("cyan_concrete", {})


class CyanConcretePowder(Block):
	def __init__(self):
		super().__init__("cyan_concrete_powder", {})


class CyanGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("cyan_glazed_terracotta", {})


class CyanShulkerBox(Block):
	def __init__(self):
		super().__init__("cyan_shulker_box", {})


class CyanStainedGlass(Block):
	def __init__(self):
		super().__init__("cyan_stained_glass", {})


class CyanStainedGlassPane(Block):
	def __init__(self):
		super().__init__("cyan_stained_glass_pane", {})


class CyanTerracotta(Block):
	def __init__(self):
		super().__init__("cyan_terracotta", {})


class CyanWool(Block):
	def __init__(self):
		super().__init__("cyan_wool", {})


class DamagedAnvil(Block):
	def __init__(self):
		super().__init__("damaged_anvil", {})


class DarkOakButton(Block):
	def __init__(self):
		super().__init__("dark_oak_button", {})


class DarkOakFence(Block):
	def __init__(self):
		super().__init__("dark_oak_fence", {})


class DarkOakFenceGate(Block):
	def __init__(self):
		super().__init__("dark_oak_fence_gate", {})


class DarkOakLeaves(Block):
	def __init__(self):
		super().__init__("dark_oak_leaves", {})


class DarkOakLog(Block):
	def __init__(self):
		super().__init__("dark_oak_log", {})


class DarkOakPlanks(Block):
	def __init__(self):
		super().__init__("dark_oak_planks", {})


class DarkOakPressurePlate(Block):
	def __init__(self):
		super().__init__("dark_oak_pressure_plate", {})


class DarkOakSign(Block):
	def __init__(self):
		super().__init__("dark_oak_sign", {})


class DarkOakSlab(Block):
	def __init__(self):
		super().__init__("dark_oak_slab", {})


class DarkOakStairs(Block):
	def __init__(self):
		super().__init__("dark_oak_stairs", {})


class DarkOakTrapdoor(Block):
	def __init__(self):
		super().__init__("dark_oak_trapdoor", {})


class DarkOakWood(Block):
	def __init__(self):
		super().__init__("dark_oak_wood", {})


class DarkPrismarine(Block):
	def __init__(self):
		super().__init__("dark_prismarine", {})


class DarkPrismarineSlab(Block):
	def __init__(self):
		super().__init__("dark_prismarine_slab", {})


class DarkPrismarineStairs(Block):
	def __init__(self):
		super().__init__("dark_prismarine_stairs", {})


class DaylightDetector(Block):
	def __init__(self):
		super().__init__("daylight_detector", {})


class DeadBrainCoralBlock(Block):
	def __init__(self):
		super().__init__("dead_brain_coral_block", {})


class DeadBubbleCoralBlock(Block):
	def __init__(self):
		super().__init__("dead_bubble_coral_block", {})


class DeadFireCoralBlock(Block):
	def __init__(self):
		super().__init__("dead_fire_coral_block", {})


class DeadHornCoralBlock(Block):
	def __init__(self):
		super().__init__("dead_horn_coral_block", {})


class DeadTubeCoralBlock(Block):
	def __init__(self):
		super().__init__("dead_tube_coral_block", {})


class DetectorRail(Block):
	def __init__(self):
		super().__init__("detector_rail", {})


class DiamondBlock(Block):
	def __init__(self):
		super().__init__("diamond_block", {})


class DiamondOre(Block):
	def __init__(self):
		super().__init__("diamond_ore", {})


class Diorite(Block):
	def __init__(self):
		super().__init__("diorite", {})


class DioriteSlab(Block):
	def __init__(self):
		super().__init__("diorite_slab", {})


class DioriteStairs(Block):
	def __init__(self):
		super().__init__("diorite_stairs", {})


class DioriteWall(Block):
	def __init__(self):
		super().__init__("diorite_wall", {})


class Dirt(Block):
	def __init__(self):
		super().__init__("dirt", {})


class Dispenser(Block):
	def __init__(self):
		super().__init__("dispenser", {})


class Dropper(Block):
	def __init__(self):
		super().__init__("dropper", {})


class DriedKelpBlock(Block):
	def __init__(self):
		super().__init__("dried_kelp_block", {})


class EmeraldBlock(Block):
	def __init__(self):
		super().__init__("emerald_block", {})


class EmeraldOre(Block):
	def __init__(self):
		super().__init__("emerald_ore", {})


class EnchantingTable(Block):
	def __init__(self):
		super().__init__("enchanting_table", {})


class EndPortalFrame(Block):
	def __init__(self):
		super().__init__("end_portal_frame", {})


class EndStone(Block):
	def __init__(self):
		super().__init__("end_stone", {})


class EndStoneBrickSlab(Block):
	def __init__(self):
		super().__init__("end_stone_brick_slab", {})


class EndStoneBrickStairs(Block):
	def __init__(self):
		super().__init__("end_stone_brick_stairs", {})


class EndStoneBrickWall(Block):
	def __init__(self):
		super().__init__("end_stone_brick_wall", {})


class EndStoneBricks(Block):
	def __init__(self):
		super().__init__("end_stone_bricks", {})


class EnderChest(Block):
	def __init__(self):
		super().__init__("ender_chest", {})


class Farmland(Block):
	def __init__(self):
		super().__init__("farmland", {})


class FireCoralBlock(Block):
	def __init__(self):
		super().__init__("fire_coral_block", {})


class FletchingTable(Block):
	def __init__(self):
		super().__init__("fletching_table", {})


class Furnace(Block):
	def __init__(self):
		super().__init__("furnace", {})


class GildedBlackstone(Block):
	def __init__(self):
		super().__init__("gilded_blackstone", {})


class Glass(Block):
	def __init__(self):
		super().__init__("glass", {})


class GlassPane(Block):
	def __init__(self):
		super().__init__("glass_pane", {})


class Glowstone(Block):
	def __init__(self):
		super().__init__("glowstone", {})


class GoldBlock(Block):
	def __init__(self):
		super().__init__("gold_block", {})


class GoldOre(Block):
	def __init__(self):
		super().__init__("gold_ore", {})


class Granite(Block):
	def __init__(self):
		super().__init__("granite", {})


class GraniteSlab(Block):
	def __init__(self):
		super().__init__("granite_slab", {})


class GraniteStairs(Block):
	def __init__(self):
		super().__init__("granite_stairs", {})


class GraniteWall(Block):
	def __init__(self):
		super().__init__("granite_wall", {})


class GrassBlock(Block):
	def __init__(self):
		super().__init__("grass_block", {})


class GrassPath(Block):
	def __init__(self):
		super().__init__("grass_path", {})


class Gravel(Block):
	def __init__(self):
		super().__init__("gravel", {})


class GrayConcrete(Block):
	def __init__(self):
		super().__init__("gray_concrete", {})


class GrayConcretePowder(Block):
	def __init__(self):
		super().__init__("gray_concrete_powder", {})


class GrayGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("gray_glazed_terracotta", {})


class GrayShulkerBox(Block):
	def __init__(self):
		super().__init__("gray_shulker_box", {})


class GrayStainedGlass(Block):
	def __init__(self):
		super().__init__("gray_stained_glass", {})


class GrayStainedGlassPane(Block):
	def __init__(self):
		super().__init__("gray_stained_glass_pane", {})


class GrayTerracotta(Block):
	def __init__(self):
		super().__init__("gray_terracotta", {})


class GrayWool(Block):
	def __init__(self):
		super().__init__("gray_wool", {})


class GreenConcrete(Block):
	def __init__(self):
		super().__init__("green_concrete", {})


class GreenConcretePowder(Block):
	def __init__(self):
		super().__init__("green_concrete_powder", {})


class GreenGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("green_glazed_terracotta", {})


class GreenShulkerBox(Block):
	def __init__(self):
		super().__init__("green_shulker_box", {})


class GreenStainedGlass(Block):
	def __init__(self):
		super().__init__("green_stained_glass", {})


class GreenStainedGlassPane(Block):
	def __init__(self):
		super().__init__("green_stained_glass_pane", {})


class GreenTerracotta(Block):
	def __init__(self):
		super().__init__("green_terracotta", {})


class GreenWool(Block):
	def __init__(self):
		super().__init__("green_wool", {})


class Grindstone(Block):
	def __init__(self):
		super().__init__("grindstone", {})


class HayBlock(Block):
	def __init__(self):
		super().__init__("hay_block", {})


class HeavyWeightedPressurePlate(Block):
	def __init__(self):
		super().__init__("heavy_weighted_pressure_plate", {})


class HoneyBlock(Block):
	def __init__(self):
		super().__init__("honey_block", {})


class HoneycombBlock(Block):
	def __init__(self):
		super().__init__("honeycomb_block", {})


class HornCoralBlock(Block):
	def __init__(self):
		super().__init__("horn_coral_block", {})


class Ice(Block):
	def __init__(self):
		super().__init__("ice", {})


class IronBars(Block):
	def __init__(self):
		super().__init__("iron_bars", {})


class IronBlock(Block):
	def __init__(self):
		super().__init__("iron_block", {})


class IronOre(Block):
	def __init__(self):
		super().__init__("iron_ore", {})


class IronTrapdoor(Block):
	def __init__(self):
		super().__init__("iron_trapdoor", {})


class JackOLantern(Block):
	def __init__(self):
		super().__init__("jack_o_lantern", {})


class Jukebox(Block):
	def __init__(self):
		super().__init__("jukebox", {})


class JungleButton(Block):
	def __init__(self):
		super().__init__("jungle_button", {})


class JungleDoor(Block):
	def __init__(self):
		super().__init__("jungle_door", {})


class JungleFence(Block):
	def __init__(self):
		super().__init__("jungle_fence", {})


class JungleFenceGate(Block):
	def __init__(self):
		super().__init__("jungle_fence_gate", {})


class JungleLeaves(Block):
	def __init__(self):
		super().__init__("jungle_leaves", {})


class JungleLog(Block):
	def __init__(self):
		super().__init__("jungle_log", {})


class JunglePlanks(Block):
	def __init__(self):
		super().__init__("jungle_planks", {})


class JunglePressurePlate(Block):
	def __init__(self):
		super().__init__("jungle_pressure_plate", {})


class JungleSlab(Block):
	def __init__(self):
		super().__init__("jungle_slab", {})


class JungleStairs(Block):
	def __init__(self):
		super().__init__("jungle_stairs", {})


class JungleTrapdoor(Block):
	def __init__(self):
		super().__init__("jungle_trapdoor", {})


class JungleWood(Block):
	def __init__(self):
		super().__init__("jungle_wood", {})


class Ladder(Block):
	def __init__(self):
		super().__init__("ladder", {})


class LapisBlock(Block):
	def __init__(self):
		super().__init__("lapis_block", {})


class LapisOre(Block):
	def __init__(self):
		super().__init__("lapis_ore", {})


class LightBlueConcrete(Block):
	def __init__(self):
		super().__init__("light_blue_concrete", {})


class LightBlueConcretePowder(Block):
	def __init__(self):
		super().__init__("light_blue_concrete_powder", {})


class LightBlueGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("light_blue_glazed_terracotta", {})


class LightBlueShulkerBox(Block):
	def __init__(self):
		super().__init__("light_blue_shulker_box", {})


class LightBlueStainedGlass(Block):
	def __init__(self):
		super().__init__("light_blue_stained_glass", {})


class LightBlueStainedGlassPane(Block):
	def __init__(self):
		super().__init__("light_blue_stained_glass_pane", {})


class LightBlueTerracotta(Block):
	def __init__(self):
		super().__init__("light_blue_terracotta", {})


class LightBlueWool(Block):
	def __init__(self):
		super().__init__("light_blue_wool", {})


class LightGrayConcrete(Block):
	def __init__(self):
		super().__init__("light_gray_concrete", {})


class LightGrayConcretePowder(Block):
	def __init__(self):
		super().__init__("light_gray_concrete_powder", {})


class LightGrayGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("light_gray_glazed_terracotta", {})


class LightGrayShulkerBox(Block):
	def __init__(self):
		super().__init__("light_gray_shulker_box", {})


class LightGrayStainedGlass(Block):
	def __init__(self):
		super().__init__("light_gray_stained_glass", {})


class LightGrayStainedGlassPane(Block):
	def __init__(self):
		super().__init__("light_gray_stained_glass_pane", {})


class LightGrayTerracotta(Block):
	def __init__(self):
		super().__init__("light_gray_terracotta", {})


class LightGrayWool(Block):
	def __init__(self):
		super().__init__("light_gray_wool", {})


class LightWeightedPressurePlate(Block):
	def __init__(self):
		super().__init__("light_weighted_pressure_plate", {})


class LimeConcrete(Block):
	def __init__(self):
		super().__init__("lime_concrete", {})


class LimeConcretePowder(Block):
	def __init__(self):
		super().__init__("lime_concrete_powder", {})


class LimeGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("lime_glazed_terracotta", {})


class LimeShulkerBox(Block):
	def __init__(self):
		super().__init__("lime_shulker_box", {})


class LimeStainedGlass(Block):
	def __init__(self):
		super().__init__("lime_stained_glass", {})


class LimeStainedGlassPane(Block):
	def __init__(self):
		super().__init__("lime_stained_glass_pane", {})


class LimeTerracotta(Block):
	def __init__(self):
		super().__init__("lime_terracotta", {})


class LimeWool(Block):
	def __init__(self):
		super().__init__("lime_wool", {})


class Lodestone(Block):
	def __init__(self):
		super().__init__("lodestone", {})


class Loom(Block):
	def __init__(self):
		super().__init__("loom", {})


class MagentaConcrete(Block):
	def __init__(self):
		super().__init__("magenta_concrete", {})


class MagentaConcretePowder(Block):
	def __init__(self):
		super().__init__("magenta_concrete_powder", {})


class MagentaGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("magenta_glazed_terracotta", {})


class MagentaShulkerBox(Block):
	def __init__(self):
		super().__init__("magenta_shulker_box", {})


class MagentaStainedGlass(Block):
	def __init__(self):
		super().__init__("magenta_stained_glass", {})


class MagentaStainedGlassPane(Block):
	def __init__(self):
		super().__init__("magenta_stained_glass_pane", {})


class MagentaTerracotta(Block):
	def __init__(self):
		super().__init__("magenta_terracotta", {})


class MagentaWool(Block):
	def __init__(self):
		super().__init__("magenta_wool", {})


class MagmaBlock(Block):
	def __init__(self):
		super().__init__("magma_block", {})


class Melon(Block):
	def __init__(self):
		super().__init__("melon", {})


class MossyCobblestone(Block):
	def __init__(self):
		super().__init__("mossy_cobblestone", {})


class MossyCobblestoneSlab(Block):
	def __init__(self):
		super().__init__("mossy_cobblestone_slab", {})


class MossyCobblestoneStairs(Block):
	def __init__(self):
		super().__init__("mossy_cobblestone_stairs", {})


class MossyCobblestoneWall(Block):
	def __init__(self):
		super().__init__("mossy_cobblestone_wall", {})


class MossyStoneBrickSlab(Block):
	def __init__(self):
		super().__init__("mossy_stone_brick_slab", {})


class MossyStoneBrickStairs(Block):
	def __init__(self):
		super().__init__("mossy_stone_brick_stairs", {})


class MossyStoneBrickWall(Block):
	def __init__(self):
		super().__init__("mossy_stone_brick_wall", {})


class MossyStoneBricks(Block):
	def __init__(self):
		super().__init__("mossy_stone_bricks", {})


class MushroomStem(Block):
	def __init__(self):
		super().__init__("mushroom_stem", {})


class Mycelium(Block):
	def __init__(self):
		super().__init__("mycelium", {})


class NetherBrickFence(Block):
	def __init__(self):
		super().__init__("nether_brick_fence", {})


class NetherBrickSlab(Block):
	def __init__(self):
		super().__init__("nether_brick_slab", {})


class NetherBrickStairs(Block):
	def __init__(self):
		super().__init__("nether_brick_stairs", {})


class NetherBrickWall(Block):
	def __init__(self):
		super().__init__("nether_brick_wall", {})


class NetherBricks(Block):
	def __init__(self):
		super().__init__("nether_bricks", {})


class NetherGoldOre(Block):
	def __init__(self):
		super().__init__("nether_gold_ore", {})


class NetherQuartzOre(Block):
	def __init__(self):
		super().__init__("nether_quartz_ore", {})


class NetherWartBlock(Block):
	def __init__(self):
		super().__init__("nether_wart_block", {})


class NetheriteBlock(Block):
	def __init__(self):
		super().__init__("netherite_block", {})


class Netherrack(Block):
	def __init__(self):
		super().__init__("netherrack", {})


class NoteBlock(Block):
	def __init__(self):
		super().__init__("note_block", {})


class OakButton(Block):
	def __init__(self):
		super().__init__("oak_button", {})


class OakFence(Block):
	def __init__(self):
		super().__init__("oak_fence", {})


class OakFenceGate(Block):
	def __init__(self):
		super().__init__("oak_fence_gate", {})


class OakLeaves(Block):
	def __init__(self):
		super().__init__("oak_leaves", {})


class OakLog(Block):
	def __init__(self):
		super().__init__("oak_log", {})


class OakPlanks(Block):
	def __init__(self):
		super().__init__("oak_planks", {})


class OakPressurePlate(Block):
	def __init__(self):
		super().__init__("oak_pressure_plate", {})


class OakSlab(Block):
	def __init__(self):
		super().__init__("oak_slab", {})


class OakStairs(Block):
	def __init__(self):
		super().__init__("oak_stairs", {})


class OakTrapdoor(Block):
	def __init__(self):
		super().__init__("oak_trapdoor", {})


class OakWood(Block):
	def __init__(self):
		super().__init__("oak_wood", {})


class Observer(Block):
	def __init__(self):
		super().__init__("observer", {})


class Obsidian(Block):
	def __init__(self):
		super().__init__("obsidian", {})


class OrangeConcrete(Block):
	def __init__(self):
		super().__init__("orange_concrete", {})


class OrangeConcretePowder(Block):
	def __init__(self):
		super().__init__("orange_concrete_powder", {})


class OrangeGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("orange_glazed_terracotta", {})


class OrangeShulkerBox(Block):
	def __init__(self):
		super().__init__("orange_shulker_box", {})


class OrangeStainedGlass(Block):
	def __init__(self):
		super().__init__("orange_stained_glass", {})


class OrangeStainedGlassPane(Block):
	def __init__(self):
		super().__init__("orange_stained_glass_pane", {})


class OrangeTerracotta(Block):
	def __init__(self):
		super().__init__("orange_terracotta", {})


class OrangeWool(Block):
	def __init__(self):
		super().__init__("orange_wool", {})


class PackedIce(Block):
	def __init__(self):
		super().__init__("packed_ice", {})


class PetrifiedOakSlab(Block):
	def __init__(self):
		super().__init__("petrified_oak_slab", {})


class PinkConcrete(Block):
	def __init__(self):
		super().__init__("pink_concrete", {})


class PinkConcretePowder(Block):
	def __init__(self):
		super().__init__("pink_concrete_powder", {})


class PinkGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("pink_glazed_terracotta", {})


class PinkShulkerBox(Block):
	def __init__(self):
		super().__init__("pink_shulker_box", {})


class PinkStainedGlass(Block):
	def __init__(self):
		super().__init__("pink_stained_glass", {})


class PinkStainedGlassPane(Block):
	def __init__(self):
		super().__init__("pink_stained_glass_pane", {})


class PinkTerracotta(Block):
	def __init__(self):
		super().__init__("pink_terracotta", {})


class PinkWool(Block):
	def __init__(self):
		super().__init__("pink_wool", {})


class Piston(Block):
	def __init__(self):
		super().__init__("piston", {})


class Podzol(Block):
	def __init__(self):
		super().__init__("podzol", {})


class PolishedAndesite(Block):
	def __init__(self):
		super().__init__("polished_andesite", {})


class PolishedAndesiteSlab(Block):
	def __init__(self):
		super().__init__("polished_andesite_slab", {})


class PolishedAndesiteStairs(Block):
	def __init__(self):
		super().__init__("polished_andesite_stairs", {})


class PolishedBasalt(Block):
	def __init__(self):
		super().__init__("polished_basalt", {})


class PolishedBlackstone(Block):
	def __init__(self):
		super().__init__("polished_blackstone", {})


class PolishedBlackstoneBrickSlab(Block):
	def __init__(self):
		super().__init__("polished_blackstone_brick_slab", {})


class PolishedBlackstoneBrickStairs(Block):
	def __init__(self):
		super().__init__("polished_blackstone_brick_stairs", {})


class PolishedBlackstoneBrickWall(Block):
	def __init__(self):
		super().__init__("polished_blackstone_brick_wall", {})


class PolishedBlackstoneBricks(Block):
	def __init__(self):
		super().__init__("polished_blackstone_bricks", {})


class PolishedBlackstoneButton(Block):
	def __init__(self):
		super().__init__("polished_blackstone_button", {})


class PolishedBlackstonePressurePlate(Block):
	def __init__(self):
		super().__init__("polished_blackstone_pressure_plate", {})


class PolishedBlackstoneSlab(Block):
	def __init__(self):
		super().__init__("polished_blackstone_slab", {})


class PolishedBlackstoneStairs(Block):
	def __init__(self):
		super().__init__("polished_blackstone_stairs", {})


class PolishedBlackstoneWall(Block):
	def __init__(self):
		super().__init__("polished_blackstone_wall", {})


class PolishedDiorite(Block):
	def __init__(self):
		super().__init__("polished_diorite", {})


class PolishedDioriteSlab(Block):
	def __init__(self):
		super().__init__("polished_diorite_slab", {})


class PolishedDioriteStairs(Block):
	def __init__(self):
		super().__init__("polished_diorite_stairs", {})


class PolishedGranite(Block):
	def __init__(self):
		super().__init__("polished_granite", {})


class PolishedGraniteSlab(Block):
	def __init__(self):
		super().__init__("polished_granite_slab", {})


class PolishedGraniteStairs(Block):
	def __init__(self):
		super().__init__("polished_granite_stairs", {})


class Prismarine(Block):
	def __init__(self):
		super().__init__("prismarine", {})


class PrismarineBrickSlab(Block):
	def __init__(self):
		super().__init__("prismarine_brick_slab", {})


class PrismarineBrickStairs(Block):
	def __init__(self):
		super().__init__("prismarine_brick_stairs", {})


class PrismarineBricks(Block):
	def __init__(self):
		super().__init__("prismarine_bricks", {})


class PrismarineSlab(Block):
	def __init__(self):
		super().__init__("prismarine_slab", {})


class PrismarineStairs(Block):
	def __init__(self):
		super().__init__("prismarine_stairs", {})


class PrismarineWall(Block):
	def __init__(self):
		super().__init__("prismarine_wall", {})


class Pumpkin(Block):
	def __init__(self):
		super().__init__("pumpkin", {})


class PurpleConcrete(Block):
	def __init__(self):
		super().__init__("purple_concrete", {})


class PurpleConcretePowder(Block):
	def __init__(self):
		super().__init__("purple_concrete_powder", {})


class PurpleGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("purple_glazed_terracotta", {})


class PurpleShulkerBox(Block):
	def __init__(self):
		super().__init__("purple_shulker_box", {})


class PurpleStainedGlass(Block):
	def __init__(self):
		super().__init__("purple_stained_glass", {})


class PurpleStainedGlassPane(Block):
	def __init__(self):
		super().__init__("purple_stained_glass_pane", {})


class PurpleTerracotta(Block):
	def __init__(self):
		super().__init__("purple_terracotta", {})


class PurpleWool(Block):
	def __init__(self):
		super().__init__("purple_wool", {})


class PurpurBlock(Block):
	def __init__(self):
		super().__init__("purpur_block", {})


class PurpurPillar(Block):
	def __init__(self):
		super().__init__("purpur_pillar", {})


class PurpurSlab(Block):
	def __init__(self):
		super().__init__("purpur_slab", {})


class PurpurStairs(Block):
	def __init__(self):
		super().__init__("purpur_stairs", {})


class QuartzBlock(Block):
	def __init__(self):
		super().__init__("quartz_block", {})


class QuartzBricks(Block):
	def __init__(self):
		super().__init__("quartz_bricks", {})


class QuartzPillar(Block):
	def __init__(self):
		super().__init__("quartz_pillar", {})


class QuartzSlab(Block):
	def __init__(self):
		super().__init__("quartz_slab", {})


class QuartzStairs(Block):
	def __init__(self):
		super().__init__("quartz_stairs", {})


class Rail(Block):
	def __init__(self):
		super().__init__("rail", {})


class RedConcrete(Block):
	def __init__(self):
		super().__init__("red_concrete", {})


class RedConcretePowder(Block):
	def __init__(self):
		super().__init__("red_concrete_powder", {})


class RedGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("red_glazed_terracotta", {})


class RedMushroomBlock(Block):
	def __init__(self):
		super().__init__("red_mushroom_block", {})


class RedNetherBrickSlab(Block):
	def __init__(self):
		super().__init__("red_nether_brick_slab", {})


class RedNetherBrickStairs(Block):
	def __init__(self):
		super().__init__("red_nether_brick_stairs", {})


class RedNetherBrickWall(Block):
	def __init__(self):
		super().__init__("red_nether_brick_wall", {})


class RedNetherBricks(Block):
	def __init__(self):
		super().__init__("red_nether_bricks", {})


class RedSand(Block):
	def __init__(self):
		super().__init__("red_sand", {})


class RedSandstone(Block):
	def __init__(self):
		super().__init__("red_sandstone", {})


class RedSandstoneSlab(Block):
	def __init__(self):
		super().__init__("red_sandstone_slab", {})


class RedSandstoneStairs(Block):
	def __init__(self):
		super().__init__("red_sandstone_stairs", {})


class RedSandstoneWall(Block):
	def __init__(self):
		super().__init__("red_sandstone_wall", {})


class RedShulkerBox(Block):
	def __init__(self):
		super().__init__("red_shulker_box", {})


class RedStainedGlass(Block):
	def __init__(self):
		super().__init__("red_stained_glass", {})


class RedStainedGlassPane(Block):
	def __init__(self):
		super().__init__("red_stained_glass_pane", {})


class RedTerracotta(Block):
	def __init__(self):
		super().__init__("red_terracotta", {})


class RedWool(Block):
	def __init__(self):
		super().__init__("red_wool", {})


class RedstoneBlock(Block):
	def __init__(self):
		super().__init__("redstone_block", {})


class RedstoneLamp(Block):
	def __init__(self):
		super().__init__("redstone_lamp", {})


class RedstoneOre(Block):
	def __init__(self):
		super().__init__("redstone_ore", {})


class RedstoneTorch(Block):
	def __init__(self):
		super().__init__("redstone_torch", {})


class Repeater(Block):
	def __init__(self):
		super().__init__("repeater", {})


class RepeatingCommandBlock(Block):
	def __init__(self):
		super().__init__("repeating_command_block", {})


class RespawnAnchor(Block):
	def __init__(self):
		super().__init__("respawn_anchor", {})


class Sand(Block):
	def __init__(self):
		super().__init__("sand", {})


class Sandstone(Block):
	def __init__(self):
		super().__init__("sandstone", {})


class SandstoneSlab(Block):
	def __init__(self):
		super().__init__("sandstone_slab", {})


class SandstoneStairs(Block):
	def __init__(self):
		super().__init__("sandstone_stairs", {})


class SandstoneWall(Block):
	def __init__(self):
		super().__init__("sandstone_wall", {})


class SeaLantern(Block):
	def __init__(self):
		super().__init__("sea_lantern", {})


class Shroomlight(Block):
	def __init__(self):
		super().__init__("shroomlight", {})


class ShulkerBox(Block):
	def __init__(self):
		super().__init__("shulker_box", {})


class SlimeBlock(Block):
	def __init__(self):
		super().__init__("slime_block", {})


class SmithingTable(Block):
	def __init__(self):
		super().__init__("smithing_table", {})


class Smoker(Block):
	def __init__(self):
		super().__init__("smoker", {})


class SmoothQuartz(Block):
	def __init__(self):
		super().__init__("smooth_quartz", {})


class SmoothQuartzSlab(Block):
	def __init__(self):
		super().__init__("smooth_quartz_slab", {})


class SmoothQuartzStairs(Block):
	def __init__(self):
		super().__init__("smooth_quartz_stairs", {})


class SmoothRedSandstone(Block):
	def __init__(self):
		super().__init__("smooth_red_sandstone", {})


class SmoothRedSandstoneSlab(Block):
	def __init__(self):
		super().__init__("smooth_red_sandstone_slab", {})


class SmoothRedSandstoneStairs(Block):
	def __init__(self):
		super().__init__("smooth_red_sandstone_stairs", {})


class SmoothSandstone(Block):
	def __init__(self):
		super().__init__("smooth_sandstone", {})


class SmoothSandstoneSlab(Block):
	def __init__(self):
		super().__init__("smooth_sandstone_slab", {})


class SmoothSandstoneStairs(Block):
	def __init__(self):
		super().__init__("smooth_sandstone_stairs", {})


class SmoothStone(Block):
	def __init__(self):
		super().__init__("smooth_stone", {})


class SmoothStoneSlab(Block):
	def __init__(self):
		super().__init__("smooth_stone_slab", {})


class Snow(Block):
	def __init__(self):
		super().__init__("snow", {})


class SnowBlock(Block):
	def __init__(self):
		super().__init__("snow_block", {})


class SoulSand(Block):
	def __init__(self):
		super().__init__("soul_sand", {})


class SoulSoil(Block):
	def __init__(self):
		super().__init__("soul_soil", {})


class SoulTorch(Block):
	def __init__(self):
		super().__init__("soul_torch", {})


class Sponge(Block):
	def __init__(self):
		super().__init__("sponge", {})


class SpruceButton(Block):
	def __init__(self):
		super().__init__("spruce_button", {})


class SpruceFence(Block):
	def __init__(self):
		super().__init__("spruce_fence", {})


class SpruceFenceGate(Block):
	def __init__(self):
		super().__init__("spruce_fence_gate", {})


class SpruceLeaves(Block):
	def __init__(self):
		super().__init__("spruce_leaves", {})


class SpruceLog(Block):
	def __init__(self):
		super().__init__("spruce_log", {})


class SprucePlanks(Block):
	def __init__(self):
		super().__init__("spruce_planks", {})


class SprucePressurePlate(Block):
	def __init__(self):
		super().__init__("spruce_pressure_plate", {})


class SpruceSlab(Block):
	def __init__(self):
		super().__init__("spruce_slab", {})


class SpruceStairs(Block):
	def __init__(self):
		super().__init__("spruce_stairs", {})


class SpruceTrapdoor(Block):
	def __init__(self):
		super().__init__("spruce_trapdoor", {})


class SpruceWood(Block):
	def __init__(self):
		super().__init__("spruce_wood", {})


class StickyPiston(Block):
	def __init__(self):
		super().__init__("sticky_piston", {})


class Stone(Block):
	def __init__(self):
		super().__init__("stone", {})


class StoneBrickSlab(Block):
	def __init__(self):
		super().__init__("stone_brick_slab", {})


class StoneBrickStairs(Block):
	def __init__(self):
		super().__init__("stone_brick_stairs", {})


class StoneBrickWall(Block):
	def __init__(self):
		super().__init__("stone_brick_wall", {})


class StoneBricks(Block):
	def __init__(self):
		super().__init__("stone_bricks", {})


class StoneButton(Block):
	def __init__(self):
		super().__init__("stone_button", {})


class StonePressurePlate(Block):
	def __init__(self):
		super().__init__("stone_pressure_plate", {})


class StoneSlab(Block):
	def __init__(self):
		super().__init__("stone_slab", {})


class StoneStairs(Block):
	def __init__(self):
		super().__init__("stone_stairs", {})


class Stonecutter(Block):
	def __init__(self):
		super().__init__("stonecutter", {})


class StrippedAcaciaLog(Block):
	def __init__(self):
		super().__init__("stripped_acacia_log", {})


class StrippedAcaciaWood(Block):
	def __init__(self):
		super().__init__("stripped_acacia_wood", {})


class StrippedBirchLog(Block):
	def __init__(self):
		super().__init__("stripped_birch_log", {})


class StrippedBirchWood(Block):
	def __init__(self):
		super().__init__("stripped_birch_wood", {})


class StrippedCrimsonHyphae(Block):
	def __init__(self):
		super().__init__("stripped_crimson_hyphae", {})


class StrippedCrimsonStem(Block):
	def __init__(self):
		super().__init__("stripped_crimson_stem", {})


class StrippedDarkOakLog(Block):
	def __init__(self):
		super().__init__("stripped_dark_oak_log", {})


class StrippedDarkOakWood(Block):
	def __init__(self):
		super().__init__("stripped_dark_oak_wood", {})


class StrippedJungleLog(Block):
	def __init__(self):
		super().__init__("stripped_jungle_log", {})


class StrippedJungleWood(Block):
	def __init__(self):
		super().__init__("stripped_jungle_wood", {})


class StrippedOakLog(Block):
	def __init__(self):
		super().__init__("stripped_oak_log", {})


class StrippedOakWood(Block):
	def __init__(self):
		super().__init__("stripped_oak_wood", {})


class StrippedSpruceLog(Block):
	def __init__(self):
		super().__init__("stripped_spruce_log", {})


class StrippedSpruceWood(Block):
	def __init__(self):
		super().__init__("stripped_spruce_wood", {})


class StrippedWarpedHyphae(Block):
	def __init__(self):
		super().__init__("stripped_warped_hyphae", {})


class StrippedWarpedStem(Block):
	def __init__(self):
		super().__init__("stripped_warped_stem", {})


class Target(Block):
	def __init__(self):
		super().__init__("target", {})


class Terracotta(Block):
	def __init__(self):
		super().__init__("terracotta", {})


class Tnt(Block):
	def __init__(self):
		super().__init__("tnt", {})


class Torch(Block):
	def __init__(self):
		super().__init__("torch", {})


class TrappedChest(Block):
	def __init__(self):
		super().__init__("trapped_chest", {})


class TubeCoralBlock(Block):
	def __init__(self):
		super().__init__("tube_coral_block", {})


class WarpedFence(Block):
	def __init__(self):
		super().__init__("warped_fence", {})


class WarpedFenceGate(Block):
	def __init__(self):
		super().__init__("warped_fence_gate", {})


class WarpedButton(Block):
	def __init__(self):
		super().__init__("warped_button", {})


class WarpedHyphae(Block):
	def __init__(self):
		super().__init__("warped_hyphae", {})


class WarpedNylium(Block):
	def __init__(self):
		super().__init__("warped_nylium", {})


class WarpedPlanks(Block):
	def __init__(self):
		super().__init__("warped_planks", {})


class WarpedPressurePlate(Block):
	def __init__(self):
		super().__init__("warped_pressure_plate", {})


class WarpedSlab(Block):
	def __init__(self):
		super().__init__("warped_slab", {})


class WarpedStairs(Block):
	def __init__(self):
		super().__init__("warped_stairs", {})


class WarpedStem(Block):
	def __init__(self):
		super().__init__("warped_stem", {})


class WarpedTrapdoor(Block):
	def __init__(self):
		super().__init__("warped_trapdoor", {})


class WarpedWartBlock(Block):
	def __init__(self):
		super().__init__("warped_wart_block", {})


class Water(Block):
	def __init__(self):
		super().__init__("water", {})


class WetSponge(Block):
	def __init__(self):
		super().__init__("wet_sponge", {})


class WhiteConcrete(Block):
	def __init__(self):
		super().__init__("white_concrete", {})


class WhiteConcretePowder(Block):
	def __init__(self):
		super().__init__("white_concrete_powder", {})


class WhiteGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("white_glazed_terracotta", {})


class WhiteShulkerBox(Block):
	def __init__(self):
		super().__init__("white_shulker_box", {})


class WhiteStainedGlass(Block):
	def __init__(self):
		super().__init__("white_stained_glass", {})


class WhiteStainedGlassPane(Block):
	def __init__(self):
		super().__init__("white_stained_glass_pane", {})


class WhiteTerracotta(Block):
	def __init__(self):
		super().__init__("white_terracotta", {})


class WhiteWool(Block):
	def __init__(self):
		super().__init__("white_wool", {})


class YellowConcrete(Block):
	def __init__(self):
		super().__init__("yellow_concrete", {})


class YellowConcretePowder(Block):
	def __init__(self):
		super().__init__("yellow_concrete_powder", {})


class YellowGlazedTerracotta(Block):
	def __init__(self):
		super().__init__("yellow_glazed_terracotta", {})


class YellowShulkerBox(Block):
	def __init__(self):
		super().__init__("yellow_shulker_box", {})


class YellowStainedGlass(Block):
	def __init__(self):
		super().__init__("yellow_stained_glass", {})


class YellowStainedGlassPane(Block):
	def __init__(self):
		super().__init__("yellow_stained_glass_pane", {})


class YellowTerracotta(Block):
	def __init__(self):
		super().__init__("yellow_terracotta", {})


class YellowWool(Block):
	def __init__(self):
		super().__init__("yellow_wool", {})


class BlockList:
	def __init__(self):
		pass

	ACACIA_BUTTON = AcaciaButton()
	ACACIA_DOOR = AcaciaDoor()
	ACACIA_FENCE = AcaciaFence()
	ACACIA_FENCE_GATE = AcaciaFenceGate()
	ACACIA_LEAVES = AcaciaLeaves()
	ACACIA_LOG = AcaciaLog()
	ACACIA_PLANKS = AcaciaPlanks()
	ACACIA_PRESSURE_PLATE = AcaciaPressurePlate()
	ACACIA_SIGN = AcaciaSign()
	ACACIA_SLAB = AcaciaSlab()
	ACACIA_STAIRS = AcaciaStairs()
	ACACIA_TRAPDOOR = AcaciaTrapdoor()
	ACACIA_WOOD = AcaciaWood()
	ACTIVATOR_RAIL = ActivatorRail()
	AIR = Air()
	ANCIENT_DEBRIS = AncientDebris()
	ANDESITE = Andesite()
	ANDESITE_SLAB = AndesiteSlab()
	ANDESITE_STAIRS = AndesiteStairs()
	ANDESITE_WALL = AndesiteWall()
	ANVIL = Anvil()
	BARREL = Barrel()
	BASALT = Basalt()
	BARRIER = Barrier()
	BEACON = Beacon()
	BEDROCK = Bedrock()
	BEE_NEST = BeeNest()
	BEEHIVE = Beehive()
	BELL = Bell()
	BIRCH_BUTTON = BirchButton()
	BIRCH_DOOR = BirchDoor()
	BIRCH_FENCE = BirchFence()
	BIRCH_FENCE_GATE = BirchFenceGate()
	BIRCH_LEAVES = BirchLeaves()
	BIRCH_LOG = BirchLog()
	BIRCH_PLANKS = BirchPlanks()
	BIRCH_PRESSURE_PLATE = BirchPressurePlate()
	BIRCH_SIGN = BirchSign()
	BIRCH_SLAB = BirchSlab()
	BIRCH_STAIRS = BirchStairs()
	BIRCH_TRAPDOOR = BirchTrapdoor()
	BIRCH_WOOD = BirchWood()
	BLACK_CONCRETE = BlackConcrete()
	BLACK_CONCRETE_POWDER = BlackConcretePowder()
	BLACK_GLAZED_TERRACOTTA = BlackGlazedTerracotta()
	BLACK_SHULKER_BOX = BlackShulkerBox()
	BLACK_STAINED_GLASS = BlackStainedGlass()
	BLACK_STAINED_GLASS_PANE = BlackStainedGlassPane()
	BLACK_TERRACOTTA = BlackTerracotta()
	BLACK_WOOL = BlackWool()
	BLACKSTONE = Blackstone()
	BLACKSTONE_SLAB = BlackstoneSlab()
	BLACKSTONE_STAIRS = BlackstoneStairs()
	BLACKSTONE_WALL = BlackstoneWall()
	BLAST_FURNACE = BlastFurnace()
	BLUE_CONCRETE = BlueConcrete()
	BLUE_CONCRETE_POWDER = BlueConcretePowder()
	BLUE_GLAZED_TERRACOTTA = BlueGlazedTerracotta()
	BLUE_ICE = BlueIce()
	BLUE_SHULKER_BOX = BlueShulkerBox()
	BLUE_STAINED_GLASS = BlueStainedGlass()
	BLUE_STAINED_GLASS_PANE = BlueStainedGlassPane()
	BLUE_TERRACOTTA = BlueTerracotta()
	BLUE_WOOL = BlueWool()
	BONE_BLOCK = BoneBlock()
	BOOKSHELF = Bookshelf()
	BRAIN_CORAL_BLOCK = BrainCoralBlock()
	BRICK_SLAB = BrickSlab()
	BRICK_STAIRS = BrickStairs()
	BRICK_WALL = BrickWall()
	BRICKS = Bricks()
	BROWN_CONCRETE = BrownConcrete()
	BROWN_CONCRETE_POWDER = BrownConcretePowder()
	BROWN_GLAZED_TERRACOTTA = BrownGlazedTerracotta()
	BROWN_MUSHROOM_BLOCK = BrownMushroomBlock()
	BROWN_SHULKER_BOX = BrownShulkerBox()
	BROWN_STAINED_GLASS = BrownStainedGlass()
	BROWN_STAINED_GLASS_PANE = BrownStainedGlassPane()
	BROWN_TERRACOTTA = BrownTerracotta()
	BROWN_WOOL = BrownWool()
	BUBBLE_CORAL_BLOCK = BubbleCoralBlock()
	CACTUS = Cactus()
	CARTOGRAPHY_TABLE = CartographyTable()
	CARVED_PUMPKIN = CarvedPumpkin()
	CAULDRON = Cauldron()
	CHAIN = Chain()
	CHAIN_COMMAND_BLOCK = ChainCommandBlock()
	CHEST = Chest()
	CHIPPED_ANVIL = ChippedAnvil()
	CHISELED_NETHER_BRICKS = ChiseledNetherBricks()
	CHISELED_POLISHED_BLACKSTONE = ChiseledPolishedBlackstone()
	CHISELED_QUARTZ_BLOCK = ChiseledQuartzBlock()
	CHISELED_RED_SANDSTONE = ChiseledRedSandstone()
	CHISELED_SANDSTONE = ChiseledSandstone()
	CHISELED_STONE_BRICKS = ChiseledStoneBricks()
	CHORUS_FLOWER = ChorusFlower()
	CHORUS_PLANT = ChorusPlant()
	CLAY = Clay()
	COAL_BLOCK = CoalBlock()
	COAL_ORE = CoalOre()
	COARSE_DIRT = CoarseDirt()
	COBBLESTONE = Cobblestone()
	COBBLESTONE_SLAB = CobblestoneSlab()
	COBBLESTONE_STAIRS = CobblestoneStairs()
	COBBLESTONE_WALL = CobblestoneWall()
	COBWEB = Cobweb()
	COMMAND_BLOCK = CommandBlock()
	COMPARATOR = Comparator()
	COMPOSTER = Composter()
	CONDUIT = Conduit()
	CRACKED_NETHER_BRICKS = CrackedNetherBricks()
	CRACKED_POLISHED_BLACKSTONE_BRICKS = CrackedPolishedBlackstoneBricks()
	CRACKED_STONE_BRICKS = CrackedStoneBricks()
	CRAFTING_TABLE = CraftingTable()
	CRIMSON_BUTTON = CrimsonButton()
	CRIMSON_DOOR = CrimsonDoor()
	CRIMSON_FENCE = CrimsonFence()
	CRIMSON_FENCE_GATE = CrimsonFenceGate()
	CRIMSON_HYPHAE = CrimsonHyphae()
	CRIMSON_NYLIUM = CrimsonNylium()
	CRIMSON_PLANKS = CrimsonPlanks()
	CRIMSON_PRESSURE_PLATE = CrimsonPressurePlate()
	CRIMSON_SIGN = CrimsonSign()
	CRIMSON_SLAB = CrimsonSlab()
	CRIMSON_STAIRS = CrimsonStairs()
	CRIMSON_STEM = CrimsonStem()
	CRIMSON_TRAPDOOR = CrimsonTrapdoor()
	CRYING_OBSIDIAN = CryingObsidian()
	CUT_RED_SANDSTONE = CutRedSandstone()
	CUT_RED_SANDSTONE_SLAB = CutRedSandstoneSlab()
	CUT_SANDSTONE = CutSandstone()
	CUT_SANDSTONE_SLAB = CutSandstoneSlab()
	CYAN_CONCRETE = CyanConcrete()
	CYAN_CONCRETE_POWDER = CyanConcretePowder()
	CYAN_GLAZED_TERRACOTTA = CyanGlazedTerracotta()
	CYAN_SHULKER_BOX = CyanShulkerBox()
	CYAN_STAINED_GLASS = CyanStainedGlass()
	CYAN_STAINED_GLASS_PANE = CyanStainedGlassPane()
	CYAN_TERRACOTTA = CyanTerracotta()
	CYAN_WOOL = CyanWool()
	DAMAGED_ANVIL = DamagedAnvil()
	DARK_OAK_BUTTON = DarkOakButton()
	DARK_OAK_FENCE = DarkOakFence()
	DARK_OAK_FENCE_GATE = DarkOakFenceGate()
	DARK_OAK_LEAVES = DarkOakLeaves()
	DARK_OAK_LOG = DarkOakLog()
	DARK_OAK_PLANKS = DarkOakPlanks()
	DARK_OAK_PRESSURE_PLATE = DarkOakPressurePlate()
	DARK_OAK_SIGN = DarkOakSign()
	DARK_OAK_SLAB = DarkOakSlab()
	DARK_OAK_STAIRS = DarkOakStairs()
	DARK_OAK_TRAPDOOR = DarkOakTrapdoor()
	DARK_OAK_WOOD = DarkOakWood()
	DARK_PRISMARINE = DarkPrismarine()
	DARK_PRISMARINE_SLAB = DarkPrismarineSlab()
	DARK_PRISMARINE_STAIRS = DarkPrismarineStairs()
	DAYLIGHT_DETECTOR = DaylightDetector()
	DEAD_BRAIN_CORAL_BLOCK = DeadBrainCoralBlock()
	DEAD_BUBBLE_CORAL_BLOCK = DeadBubbleCoralBlock()
	DEAD_FIRE_CORAL_BLOCK = DeadFireCoralBlock()
	DEAD_HORN_CORAL_BLOCK = DeadHornCoralBlock()
	DEAD_TUBE_CORAL_BLOCK = DeadTubeCoralBlock()
	DETECTOR_RAIL = DetectorRail()
	DIAMOND_BLOCK = DiamondBlock()
	DIAMOND_ORE = DiamondOre()
	DIORITE = Diorite()
	DIORITE_SLAB = DioriteSlab()
	DIORITE_STAIRS = DioriteStairs()
	DIORITE_WALL = DioriteWall()
	DIRT = Dirt()
	DISPENSER = Dispenser()
	DROPPER = Dropper()
	DRIED_KELP_BLOCK = DriedKelpBlock()
	EMERALD_BLOCK = EmeraldBlock()
	EMERALD_ORE = EmeraldOre()
	ENCHANTING_TABLE = EnchantingTable()
	END_PORTAL_FRAME = EndPortalFrame()
	END_STONE = EndStone()
	END_STONE_BRICK_SLAB = EndStoneBrickSlab()
	END_STONE_BRICK_STAIRS = EndStoneBrickStairs()
	END_STONE_BRICK_WALL = EndStoneBrickWall()
	END_STONE_BRICKS = EndStoneBricks()
	ENDER_CHEST = EnderChest()
	FARMLAND = Farmland()
	FIRE_CORAL_BLOCK = FireCoralBlock()
	FLETCHING_TABLE = FletchingTable()
	FURNACE = Furnace()
	GILDED_BLACKSTONE = GildedBlackstone()
	GLASS = Glass()
	GLASS_PANE = GlassPane()
	GLOWSTONE = Glowstone()
	GOLD_BLOCK = GoldBlock()
	GOLD_ORE = GoldOre()
	GRANITE = Granite()
	GRANITE_SLAB = GraniteSlab()
	GRANITE_STAIRS = GraniteStairs()
	GRANITE_WALL = GraniteWall()
	GRASS_BLOCK = GrassBlock()
	GRASS_PATH = GrassPath()
	GRAVEL = Gravel()
	GRAY_CONCRETE = GrayConcrete()
	GRAY_CONCRETE_POWDER = GrayConcretePowder()
	GRAY_GLAZED_TERRACOTTA = GrayGlazedTerracotta()
	GRAY_SHULKER_BOX = GrayShulkerBox()
	GRAY_STAINED_GLASS = GrayStainedGlass()
	GRAY_STAINED_GLASS_PANE = GrayStainedGlassPane()
	GRAY_TERRACOTTA = GrayTerracotta()
	GRAY_WOOL = GrayWool()
	GREEN_CONCRETE = GreenConcrete()
	GREEN_CONCRETE_POWDER = GreenConcretePowder()
	GREEN_GLAZED_TERRACOTTA = GreenGlazedTerracotta()
	GREEN_SHULKER_BOX = GreenShulkerBox()
	GREEN_STAINED_GLASS = GreenStainedGlass()
	GREEN_STAINED_GLASS_PANE = GreenStainedGlassPane()
	GREEN_TERRACOTTA = GreenTerracotta()
	GREEN_WOOL = GreenWool()
	GRINDSTONE = Grindstone()
	HAY_BLOCK = HayBlock()
	HEAVY_WEIGHTED_PRESSURE_PLATE = HeavyWeightedPressurePlate()
	HONEY_BLOCK = HoneyBlock()
	HONEYCOMB_BLOCK = HoneycombBlock()
	HORN_CORAL_BLOCK = HornCoralBlock()
	ICE = Ice()
	IRON_BARS = IronBars()
	IRON_BLOCK = IronBlock()
	IRON_ORE = IronOre()
	IRON_TRAPDOOR = IronTrapdoor()
	JACK_O_LANTERN = JackOLantern()
	JUKEBOX = Jukebox()
	JUNGLE_BUTTON = JungleButton()
	JUNGLE_DOOR = JungleDoor()
	JUNGLE_FENCE = JungleFence()
	JUNGLE_FENCE_GATE = JungleFenceGate()
	JUNGLE_LEAVES = JungleLeaves()
	JUNGLE_LOG = JungleLog()
	JUNGLE_PLANKS = JunglePlanks()
	JUNGLE_PRESSURE_PLATE = JunglePressurePlate()
	JUNGLE_SLAB = JungleSlab()
	JUNGLE_STAIRS = JungleStairs()
	JUNGLE_TRAPDOOR = JungleTrapdoor()
	JUNGLE_WOOD = JungleWood()
	LADDER = Ladder()
	LAPIS_BLOCK = LapisBlock()
	LAPIS_ORE = LapisOre()
	LIGHT_BLUE_CONCRETE = LightBlueConcrete()
	LIGHT_BLUE_CONCRETE_POWDER = LightBlueConcretePowder()
	LIGHT_BLUE_GLAZED_TERRACOTTA = LightBlueGlazedTerracotta()
	LIGHT_BLUE_SHULKER_BOX = LightBlueShulkerBox()
	LIGHT_BLUE_STAINED_GLASS = LightBlueStainedGlass()
	LIGHT_BLUE_STAINED_GLASS_PANE = LightBlueStainedGlassPane()
	LIGHT_BLUE_TERRACOTTA = LightBlueTerracotta()
	LIGHT_BLUE_WOOL = LightBlueWool()
	LIGHT_GRAY_CONCRETE = LightGrayConcrete()
	LIGHT_GRAY_CONCRETE_POWDER = LightGrayConcretePowder()
	LIGHT_GRAY_GLAZED_TERRACOTTA = LightGrayGlazedTerracotta()
	LIGHT_GRAY_SHULKER_BOX = LightGrayShulkerBox()
	LIGHT_GRAY_STAINED_GLASS = LightGrayStainedGlass()
	LIGHT_GRAY_STAINED_GLASS_PANE = LightGrayStainedGlassPane()
	LIGHT_GRAY_TERRACOTTA = LightGrayTerracotta()
	LIGHT_GRAY_WOOL = LightGrayWool()
	LIGHT_WEIGHTED_PRESSURE_PLATE = LightWeightedPressurePlate()
	LIME_CONCRETE = LimeConcrete()
	LIME_CONCRETE_POWDER = LimeConcretePowder()
	LIME_GLAZED_TERRACOTTA = LimeGlazedTerracotta()
	LIME_SHULKER_BOX = LimeShulkerBox()
	LIME_STAINED_GLASS = LimeStainedGlass()
	LIME_STAINED_GLASS_PANE = LimeStainedGlassPane()
	LIME_TERRACOTTA = LimeTerracotta()
	LIME_WOOL = LimeWool()
	LODESTONE = Lodestone()
	LOOM = Loom()
	MAGENTA_CONCRETE = MagentaConcrete()
	MAGENTA_CONCRETE_POWDER = MagentaConcretePowder()
	MAGENTA_GLAZED_TERRACOTTA = MagentaGlazedTerracotta()
	MAGENTA_SHULKER_BOX = MagentaShulkerBox()
	MAGENTA_STAINED_GLASS = MagentaStainedGlass()
	MAGENTA_STAINED_GLASS_PANE = MagentaStainedGlassPane()
	MAGENTA_TERRACOTTA = MagentaTerracotta()
	MAGENTA_WOOL = MagentaWool()
	MAGMA_BLOCK = MagmaBlock()
	MELON = Melon()
	MOSSY_COBBLESTONE = MossyCobblestone()
	MOSSY_COBBLESTONE_SLAB = MossyCobblestoneSlab()
	MOSSY_COBBLESTONE_STAIRS = MossyCobblestoneStairs()
	MOSSY_COBBLESTONE_WALL = MossyCobblestoneWall()
	MOSSY_STONE_BRICK_SLAB = MossyStoneBrickSlab()
	MOSSY_STONE_BRICK_STAIRS = MossyStoneBrickStairs()
	MOSSY_STONE_BRICK_WALL = MossyStoneBrickWall()
	MOSSY_STONE_BRICKS = MossyStoneBricks()
	MUSHROOM_STEM = MushroomStem()
	MYCELIUM = Mycelium()
	NETHER_BRICK_FENCE = NetherBrickFence()
	NETHER_BRICK_SLAB = NetherBrickSlab()
	NETHER_BRICK_STAIRS = NetherBrickStairs()
	NETHER_BRICK_WALL = NetherBrickWall()
	NETHER_BRICKS = NetherBricks()
	NETHER_GOLD_ORE = NetherGoldOre()
	NETHER_QUARTZ_ORE = NetherQuartzOre()
	NETHER_WART_BLOCK = NetherWartBlock()
	NETHERITE_BLOCK = NetheriteBlock()
	NETHERRACK = Netherrack()
	NOTE_BLOCK = NoteBlock()
	OAK_BUTTON = OakButton()
	OAK_FENCE = OakFence()
	OAK_FENCE_GATE = OakFenceGate()
	OAK_LEAVES = OakLeaves()
	OAK_LOG = OakLog()
	OAK_PLANKS = OakPlanks()
	OAK_PRESSURE_PLATE = OakPressurePlate()
	OAK_SLAB = OakSlab()
	OAK_STAIRS = OakStairs()
	OAK_TRAPDOOR = OakTrapdoor()
	OAK_WOOD = OakWood()
	OBSERVER = Observer()
	OBSIDIAN = Obsidian()
	ORANGE_CONCRETE = OrangeConcrete()
	ORANGE_CONCRETE_POWDER = OrangeConcretePowder()
	ORANGE_GLAZED_TERRACOTTA = OrangeGlazedTerracotta()
	ORANGE_SHULKER_BOX = OrangeShulkerBox()
	ORANGE_STAINED_GLASS = OrangeStainedGlass()
	ORANGE_STAINED_GLASS_PANE = OrangeStainedGlassPane()
	ORANGE_TERRACOTTA = OrangeTerracotta()
	ORANGE_WOOL = OrangeWool()
	PACKED_ICE = PackedIce()
	PETRIFIED_OAK_SLAB = PetrifiedOakSlab()
	PINK_CONCRETE = PinkConcrete()
	PINK_CONCRETE_POWDER = PinkConcretePowder()
	PINK_GLAZED_TERRACOTTA = PinkGlazedTerracotta()
	PINK_SHULKER_BOX = PinkShulkerBox()
	PINK_STAINED_GLASS = PinkStainedGlass()
	PINK_STAINED_GLASS_PANE = PinkStainedGlassPane()
	PINK_TERRACOTTA = PinkTerracotta()
	PINK_WOOL = PinkWool()
	PISTON = Piston()
	PODZOL = Podzol()
	POLISHED_ANDESITE = PolishedAndesite()
	POLISHED_ANDESITE_SLAB = PolishedAndesiteSlab()
	POLISHED_ANDESITE_STAIRS = PolishedAndesiteStairs()
	POLISHED_BASALT = PolishedBasalt()
	POLISHED_BLACKSTONE = PolishedBlackstone()
	POLISHED_BLACKSTONE_BRICK_SLAB = PolishedBlackstoneBrickSlab()
	POLISHED_BLACKSTONE_BRICK_STAIRS = PolishedBlackstoneBrickStairs()
	POLISHED_BLACKSTONE_BRICK_WALL = PolishedBlackstoneBrickWall()
	POLISHED_BLACKSTONE_BRICKS = PolishedBlackstoneBricks()
	POLISHED_BLACKSTONE_BUTTON = PolishedBlackstoneButton()
	POLISHED_BLACKSTONE_PRESSURE_PLATE = PolishedBlackstonePressurePlate()
	POLISHED_BLACKSTONE_SLAB = PolishedBlackstoneSlab()
	POLISHED_BLACKSTONE_STAIRS = PolishedBlackstoneStairs()
	POLISHED_BLACKSTONE_WALL = PolishedBlackstoneWall()
	POLISHED_DIORITE = PolishedDiorite()
	POLISHED_DIORITE_SLAB = PolishedDioriteSlab()
	POLISHED_DIORITE_STAIRS = PolishedDioriteStairs()
	POLISHED_GRANITE = PolishedGranite()
	POLISHED_GRANITE_SLAB = PolishedGraniteSlab()
	POLISHED_GRANITE_STAIRS = PolishedGraniteStairs()
	PRISMARINE = Prismarine()
	PRISMARINE_BRICK_SLAB = PrismarineBrickSlab()
	PRISMARINE_BRICK_STAIRS = PrismarineBrickStairs()
	PRISMARINE_BRICKS = PrismarineBricks()
	PRISMARINE_SLAB = PrismarineSlab()
	PRISMARINE_STAIRS = PrismarineStairs()
	PRISMARINE_WALL = PrismarineWall()
	PUMPKIN = Pumpkin()
	PURPLE_CONCRETE = PurpleConcrete()
	PURPLE_CONCRETE_POWDER = PurpleConcretePowder()
	PURPLE_GLAZED_TERRACOTTA = PurpleGlazedTerracotta()
	PURPLE_SHULKER_BOX = PurpleShulkerBox()
	PURPLE_STAINED_GLASS = PurpleStainedGlass()
	PURPLE_STAINED_GLASS_PANE = PurpleStainedGlassPane()
	PURPLE_TERRACOTTA = PurpleTerracotta()
	PURPLE_WOOL = PurpleWool()
	PURPUR_BLOCK = PurpurBlock()
	PURPUR_PILLAR = PurpurPillar()
	PURPUR_SLAB = PurpurSlab()
	PURPUR_STAIRS = PurpurStairs()
	QUARTZ_BLOCK = QuartzBlock()
	QUARTZ_BRICKS = QuartzBricks()
	QUARTZ_PILLAR = QuartzPillar()
	QUARTZ_SLAB = QuartzSlab()
	QUARTZ_STAIRS = QuartzStairs()
	RAIL = Rail()
	RED_CONCRETE = RedConcrete()
	RED_CONCRETE_POWDER = RedConcretePowder()
	RED_GLAZED_TERRACOTTA = RedGlazedTerracotta()
	RED_MUSHROOM_BLOCK = RedMushroomBlock()
	RED_NETHER_BRICK_SLAB = RedNetherBrickSlab()
	RED_NETHER_BRICK_STAIRS = RedNetherBrickStairs()
	RED_NETHER_BRICK_WALL = RedNetherBrickWall()
	RED_NETHER_BRICKS = RedNetherBricks()
	RED_SAND = RedSand()
	RED_SANDSTONE = RedSandstone()
	RED_SANDSTONE_SLAB = RedSandstoneSlab()
	RED_SANDSTONE_STAIRS = RedSandstoneStairs()
	RED_SANDSTONE_WALL = RedSandstoneWall()
	RED_SHULKER_BOX = RedShulkerBox()
	RED_STAINED_GLASS = RedStainedGlass()
	RED_STAINED_GLASS_PANE = RedStainedGlassPane()
	RED_TERRACOTTA = RedTerracotta()
	RED_WOOL = RedWool()
	REDSTONE_BLOCK = RedstoneBlock()
	REDSTONE_LAMP = RedstoneLamp()
	REDSTONE_ORE = RedstoneOre()
	REDSTONE_TORCH = RedstoneTorch()
	REPEATER = Repeater()
	REPEATING_COMMAND_BLOCK = RepeatingCommandBlock()
	RESPAWN_ANCHOR = RespawnAnchor()
	SAND = Sand()
	SANDSTONE = Sandstone()
	SANDSTONE_SLAB = SandstoneSlab()
	SANDSTONE_STAIRS = SandstoneStairs()
	SANDSTONE_WALL = SandstoneWall()
	SEA_LANTERN = SeaLantern()
	SHROOMLIGHT = Shroomlight()
	SHULKER_BOX = ShulkerBox()
	SLIME_BLOCK = SlimeBlock()
	SMITHING_TABLE = SmithingTable()
	SMOKER = Smoker()
	SMOOTH_QUARTZ = SmoothQuartz()
	SMOOTH_QUARTZ_SLAB = SmoothQuartzSlab()
	SMOOTH_QUARTZ_STAIRS = SmoothQuartzStairs()
	SMOOTH_RED_SANDSTONE = SmoothRedSandstone()
	SMOOTH_RED_SANDSTONE_SLAB = SmoothRedSandstoneSlab()
	SMOOTH_RED_SANDSTONE_STAIRS = SmoothRedSandstoneStairs()
	SMOOTH_SANDSTONE = SmoothSandstone()
	SMOOTH_SANDSTONE_SLAB = SmoothSandstoneSlab()
	SMOOTH_SANDSTONE_STAIRS = SmoothSandstoneStairs()
	SMOOTH_STONE = SmoothStone()
	SMOOTH_STONE_SLAB = SmoothStoneSlab()
	SNOW = Snow()
	SNOW_BLOCK = SnowBlock()
	SOUL_SAND = SoulSand()
	SOUL_SOIL = SoulSoil()
	SOUL_TORCH = SoulTorch()
	SPONGE = Sponge()
	SPRUCE_BUTTON = SpruceButton()
	SPRUCE_FENCE = SpruceFence()
	SPRUCE_FENCE_GATE = SpruceFenceGate()
	SPRUCE_LEAVES = SpruceLeaves()
	SPRUCE_LOG = SpruceLog()
	SPRUCE_PLANKS = SprucePlanks()
	SPRUCE_PRESSURE_PLATE = SprucePressurePlate()
	SPRUCE_SLAB = SpruceSlab()
	SPRUCE_STAIRS = SpruceStairs()
	SPRUCE_TRAPDOOR = SpruceTrapdoor()
	SPRUCE_WOOD = SpruceWood()
	STICKY_PISTON = StickyPiston()
	STONE = Stone()
	STONE_BRICK_SLAB = StoneBrickSlab()
	STONE_BRICK_STAIRS = StoneBrickStairs()
	STONE_BRICK_WALL = StoneBrickWall()
	STONE_BRICKS = StoneBricks()
	STONE_BUTTON = StoneButton()
	STONE_PRESSURE_PLATE = StonePressurePlate()
	STONE_SLAB = StoneSlab()
	STONE_STAIRS = StoneStairs()
	STONECUTTER = Stonecutter()
	STRIPPED_ACACIA_LOG = StrippedAcaciaLog()
	STRIPPED_ACACIA_WOOD = StrippedAcaciaWood()
	STRIPPED_BIRCH_LOG = StrippedBirchLog()
	STRIPPED_BIRCH_WOOD = StrippedBirchWood()
	STRIPPED_CRIMSON_HYPHAE = StrippedCrimsonHyphae()
	STRIPPED_CRIMSON_STEM = StrippedCrimsonStem()
	STRIPPED_DARK_OAK_LOG = StrippedDarkOakLog()
	STRIPPED_DARK_OAK_WOOD = StrippedDarkOakWood()
	STRIPPED_JUNGLE_LOG = StrippedJungleLog()
	STRIPPED_JUNGLE_WOOD = StrippedJungleWood()
	STRIPPED_OAK_LOG = StrippedOakLog()
	STRIPPED_OAK_WOOD = StrippedOakWood()
	STRIPPED_SPRUCE_LOG = StrippedSpruceLog()
	STRIPPED_SPRUCE_WOOD = StrippedSpruceWood()
	STRIPPED_WARPED_HYPHAE = StrippedWarpedHyphae()
	STRIPPED_WARPED_STEM = StrippedWarpedStem()
	TARGET = Target()
	TERRACOTTA = Terracotta()
	TNT = Tnt()
	TORCH = Torch()
	TRAPPED_CHEST = TrappedChest()
	TUBE_CORAL_BLOCK = TubeCoralBlock()
	WARPED_FENCE = WarpedFence()
	WARPED_FENCE_GATE = WarpedFenceGate()
	WARPED_BUTTON = WarpedButton()
	WARPED_HYPHAE = WarpedHyphae()
	WARPED_NYLIUM = WarpedNylium()
	WARPED_PLANKS = WarpedPlanks()
	WARPED_PRESSURE_PLATE = WarpedPressurePlate()
	WARPED_SLAB = WarpedSlab()
	WARPED_STAIRS = WarpedStairs()
	WARPED_STEM = WarpedStem()
	WARPED_TRAPDOOR = WarpedTrapdoor()
	WARPED_WART_BLOCK = WarpedWartBlock()
	WATER = Water()
	WET_SPONGE = WetSponge()
	WHITE_CONCRETE = WhiteConcrete()
	WHITE_CONCRETE_POWDER = WhiteConcretePowder()
	WHITE_GLAZED_TERRACOTTA = WhiteGlazedTerracotta()
	WHITE_SHULKER_BOX = WhiteShulkerBox()
	WHITE_STAINED_GLASS = WhiteStainedGlass()
	WHITE_STAINED_GLASS_PANE = WhiteStainedGlassPane()
	WHITE_TERRACOTTA = WhiteTerracotta()
	WHITE_WOOL = WhiteWool()
	YELLOW_CONCRETE = YellowConcrete()
	YELLOW_CONCRETE_POWDER = YellowConcretePowder()
	YELLOW_GLAZED_TERRACOTTA = YellowGlazedTerracotta()
	YELLOW_SHULKER_BOX = YellowShulkerBox()
	YELLOW_STAINED_GLASS = YellowStainedGlass()
	YELLOW_STAINED_GLASS_PANE = YellowStainedGlassPane()
	YELLOW_TERRACOTTA = YellowTerracotta()
	YELLOW_WOOL = YellowWool()



class Structure:
	"""
	Creates a user-defined structure. Allows users to add sub-structures
	and blocks, along with logic.
	"""

	def __init__(self, x, y, z):
		"""
		Create the structure. Takes the position as parameters.

		:param x: the x position of the structure
		:param y: the y position of the structure
		:param z: the z position of the structure
		"""

		self.x = x
		self.y = y
		self.z = z
		self.blocks = {}

	
	def create(self):
		"""
		The default create function for the structure.
		"""

		pass


	def set_block(self, x, y, z, block):
		"""
		Set a single block in the structure.

		:param x: the block's x position
		:param y: the block's y position
		:param z: the block's z position
		:param block: the type of block to set
		"""

		# check if the block has the required properties
		if isinstance(block, dict):
			if not "name" in block or not "properties" in block:
				raise ValueError("Block does not have reqired values")
		else:
			try:
				properties = dir(block)
				if not "name" in properties or not "properties" in properties:
					raise ValueError("Block does not have required values")
			except Exception as err:
				raise ValueError("Block does not have required values")

		# find the position
		r_x = x
		r_y = y
		r_z = z

		# add the block to the structure's block dictionary
		if not r_x in self.blocks:
			self.blocks[r_x] = {}

		if not r_y in self.blocks[r_x]:
			self.blocks[r_x][r_y] = {}

		if not r_z in self.blocks[r_x][r_y]:
			self.blocks[r_x][r_y][r_z] = None

		self.blocks[r_x][r_y][r_z] = block

	
	def add_structure(self, structure, x, y, z, rotation):
		"""
		Add a sub-structure's blocks to this structure. Returns nothing.

		:param structure: the structure to add, must have create function
		:param x: the x position to add the structure at
		:param y: the y position to add the structure at
		:param z: the z position to add the structure at
		:param rotation: the rotation of the structure, in radians
		"""

		# create an instance of the structure
		ins = structure(x, y, z)
		ins.create(rotation)

		# go through all the blocks in the instance
		for b_x in ins.blocks:
			for b_y in ins.blocks[x]:
				for b_z in ins.blocks[x][y]:
					# find the position relative to this position
					adjusted_x = x
					adjusted_y = y
					adjusted_z = z

					if ins.blocks[x][y][z] is None:
						continue

					# place the block in the self.blocks dictionary
					if not adjusted_x in self.blocks:
						self.blocks[adjusted_x] = {}

					if not adjusted_y in self.blocks[adjusted_x]:
						self.blocks[adjusted_x][adjusted_y] = {}

					self.blocks[adjusted_x][adjusted_y][adjusted_z] = \
						ins.blocks[b_x][b_y][b_z]



	def place(self, rotation, world):
		"""
		Places the structure. Takes rotation and world as parameters.

		:param rotation: the rotation to place the structure at, in radians
		:param world: the world class to use
		"""

		for x in self.blocks:
			for y in self.blocks[x]:
				for z in self.blocks[x][y]:
					if self.blocks[x][y][z] is None:
						continue

					# find the new position after rotating
					rotated_x = x
					rotated_z = z

					# use the world to set the block
					world.set_block(
						rotated_x,
						y,
						rotated_z,
						self.blocks[x][y][z]
					)




import time
import os
import datetime

from .option import Option
from random import randint
from .nbt import *
from .block import BlockList
from .shape import ShapeList
from .gamerules import GameRules
from .region import Region


class Logger:
	def log(self, message):
		print(f"{datetime.datetime.now().strftime('%m-%d-%y %H:%M:%S')} - {message}")


def cmod(n, base):
	"""
	Compute the modulus of a number like in languages such as C.
	
	:param n: the number to modulus
	:param base: the divisor
	"""

	return n - int(n / base) * base


class World:
	"""
	Creates a world. Allows users to set options for the world, set blocks,
	save the world to a folder, etc.
	"""

	def __init__(self, options=None) -> None:
		"""
		Creates the world.

		:param options: a dictionary which holds the world options.
		"""

		if options is None:
			options = {}

		self.gamerules = GameRules()
		self.options = {}
		self.world_loaded = False
		self.world_edited = False
		self.loaded_regions = {}
		self.seed = randint(1e8, 1e10)
		self.set_options(options)
		self.logger = Logger()
		self.regions = []
		self.blocks = BlockList
		self.shape = ShapeList(self)


	def set_block(self, x: int, y: int, z: int, block_type) -> bool:
		"""
		Set a single block. Returns true if the block was set, false otherwise.

		:param x: the x position of the block
		:param y: the y position of the block
		:param z: the z position of the block
		:param block_type: the block to set
		"""

		world_edited = True

		# check if position is out of bounds
		if y < 0 or y > 255:
			raise ValueError("Given position is out of world boundaries")

		# find the region coordinates
		regionX = x >> 9
		regionZ = z >> 9

		blockX = self.get_region_coord(x)
		blockZ = self.get_region_coord(z)

		# create the region if it doesn't already exist
		block_region = self.region_exists(regionX, regionZ)
		if not block_region:
			if self.world_loaded:
				region_path = f"r.{regionX}.{regionZ}.mca"
				if not region_path in self.loaded_regions:
					raise ValueError(
						"Given position is outside of loaded world boundaries"
					)

				region = Region(regionX, regionZ)
				region.load(self.loaded_regions[region_path])

			self.regions.append(Region(regionX, regionZ))
			block_region = self.regions[-1]

		# call the region set_block function
		return block_region.set_block(blockX, y, blockZ, block_type)


	def set_game_rule(self, name, value):
		"""
		Sets a gamerule.

		:param name: the name of the game rule
		:param value: the new value of the rule
		"""

		self.world_edited = True
		self.gamerules.set_rule(name, value)
	

	def get_block(self, x: int, y: int, z: int):
		"""
		Get the type of one block of the world.

		:param x: the x position of the block to get
		:param y: the y postiion of the block to get
		:param z: the z position of the block to get
		"""

		# check if the position is out of bounds
		if y < 0 or y > 255:
			raise ValueError("Given position is out of world boundaries")

		# find the coordinates of the region it is in
		regionX = x >> 9
		regionZ = z >> 9

		blockX = self.get_region_coord(x)
		blockZ = self.get_region_coord(z)

		# check if the region exists
		region = self.region_exists(regionX, regionZ)
		if not region:
			return self.blocks.AIR.get_block()

		return region.get_block(blockX, y, blockZ)


	def get_region_coord(self, pos):
		rc = cmod(pos, 512)

		return rc + 512 if rc < 0 else rc


	def save(self, folder_name: str) -> bool:
		"""
		Save the world to a folder. Returns true if successful, false otherwise.
		The files in the folder will be overwritten.

		Will create:
			- level.dat
			- region files
			- session.lock
		
		:param folder_name: the name of the folder to write the world to
		"""

		# check if the given folder name is the name of a file
		if os.path.isfile(folder_name):
			raise ValueError("Folder name is the name of a file")

		# check if the folder already exists
		if os.path.exists(folder_name):
			raise ValueError("Folder already exists")

		# set the world name
		self.options["LevelName"] = Option("LevelName", TAG_String, \
			folder_name.split("/")[-1])

		# create the folder
		os.mkdir(folder_name)
		self.logger.log("world folder created")

		# create the region folder
		os.mkdir(f"{folder_name}/region")

		# create level.dat
		self.create_level_dat(f"{folder_name}/level.dat")
		self.logger.log("level.dat created")

		# create session.lock
		with open(f"{folder_name}/session.lock", "wb") as f:
			f.write(int(time.time() * 1000).to_bytes(8, "big"))

		# save every region
		for region in self.regions:
			filename = f"{folder_name}/region/r.{region.x}.{region.z}.mca"
			region.save(filename)
			self.logger.log(f"region file r.{region.x}.{region.z}.mca created")

		self.logger.log("all region files created, world is ready")

		return True


	def set_options(self, options: dict) -> None:
		"""
		Sets default options so the user doesn't have to worry about them.

		These options will be written to the level.dat file when the world
		is saved. Takes the user-defined arguments as a parameter.

		:param options: the options the user provides.
		"""

		# option presets
		self.options["GameType"] = Option("GameType", TAG_Byte, 1)
		self.options["LastPlayed"] = Option("LastPlayed", TAG_Long, \
				int(time.time() * 1000))
		self.options["LevelName"] = Option("LevelName", TAG_String, "world")
		self.options["MapFeatures"] = Option("MapFeatures", TAG_Byte, 0)
		self.options["RandomSeed"] = Option("RandomSeed", TAG_Long, self.seed)
		self.options["SpawnX"] = Option("SpawnX", TAG_Int, 0)
		self.options["SpawnY"] = Option("SpawnY", TAG_Int, 0)
		self.options["SpawnZ"] = Option("SpawnZ", TAG_Int, 0)
		self.options["allowCommands"] = Option("allowCommands", TAG_Byte, 1)
		self.options["version"] = Option("version", TAG_Int, 19133)
		self.options["initialized"] = Option("initialized", TAG_Byte, 1)
		self.options["DataVersion"] = Option("DataVersion", TAG_Int, 2586)

		# sets the options defined by the user
		for key in options:
			if not self.option_valid(key):
				raise ValueError(f"Option not valid: {key}")
			
			self.options[key].value = options[key]


	def option_valid(self, key: str) -> bool:
		"""
		Checks if an option is a possible option.
		Returns true if valid, false otherwise.

		:param key: the name of the option
		"""

		return key in Option.VALID_OPTIONS
		

	def create_level_dat(self, file_path: str) -> None:
		"""
		Creates the level.dat file, using an NBT library.

		:param file_path: the file path to write to.
		"""

		# create the nbt file
		nbtfile = NBTFile()
		
		# create the root tag
		root_tag = TAG_Compound()
		root_tag.name = "Data"

		# add the options
		for op in self.options:
			root_tag.tags.append(self.options[op].create_tag())

		# create the world generator, to make a void world
		root_tag.tags.append(self.create_generator())
		
		# create data pack tag
		data_pack_tag = TAG_Compound()
		data_pack_tag.name = "DataPacks"

		disabled_list = TAG_List(name="Disabled", type=TAG_String)
		enabled_list = TAG_List(name="Enabled", type=TAG_String)
		enabled_list.tags.append(TAG_String(value="vanilla"))

		data_pack_tag.tags.extend([disabled_list, enabled_list])
		root_tag.tags.append(data_pack_tag)

		# create dragon fight tag
		dragon_fight = TAG_Compound()
		dragon_fight.name = "DragonFight"

		gateway_tag = TAG_List(name="Gateways", type=TAG_Int)

		for i in range(20):
			gateway_tag.tags.append(TAG_Int(value=i))

		dragon_fight.tags.extend([
			TAG_Byte(name="DragonKilled", value=1),
			TAG_Byte(name="PreviouslyKilled", value=1),
			gateway_tag
		])

		root_tag.tags.append(dragon_fight)

		# add the gamerules to the level.dat
		gamerule_tag = TAG_Compound()
		gamerule_tag.name = "GameRules"

		for rule in self.gamerules.rules:
			str_tag = TAG_String(name=rule["name"], value=rule["value"])
			gamerule_tag.tags.append(str_tag)

		root_tag.tags.append(gamerule_tag)

		nbtfile.tags.append(root_tag)
		nbtfile.write_file(file_path)

	
	def region_exists(self, x, z):
		"""
		Checks if a region exists at the given coordinates.

		:param x: the x position to check
		:param z: the z position to check
		"""

		for region in self.regions:
			if region.x == x and region.z == z:
				return region

		return False


	def load(self, world_folder):
		"""
		Load a world from a given folder.
		Loads each region file as needed, overwriting the original file.

		:param world_folder: the path to the folder containing the world.
		"""

		# check if the level.dat exists
		level_dat = os.path.join(world_folder, "level.dat")
		if not os.path.exists(level_dat):
			raise ValueError(f"Cannot find file at path {level_dat}")

		# check if the session.lock exists
		session_lock = os.path.join(world_folder, "session.lock")
		if not os.path.exists(session_lock):
			raise ValueError(f"Cannot find file at path {session_lock}")

		# check if the world region data exists
		region_dir = os.path.join(world_folder, "region")
		if not os.path.exists(region_dir):
			raise ValueError(f"Cannot find folder at path {region_dir}")

		# check if region files are present
		region_list = os.listdir(region_dir)
		if len(region_list) == 0:
			raise ValueError(f"No region files found in world folder")

		self.world_loaded = True
		self.loaded_regions = {
			v: os.path.join(region_dir, v) for v in region_list
		}


	def create_generator(self):
		"""
		Creates the generator options for the level.dat.
		Takes no arguments, returns TAG_Compound.

		Format:

		WorldGenSettings: TAG_Compound
			dimensions: TAG_Compound
				minecraft:overworld: TAG_Compound
					generator: TAG_Compound
						settings: TAG_Compound
							structures: TAG_Compound
								structures: TAG_Compound
									minecraft:village: TAG_Compound
										salt: TAG_Int
										separation: TAG_Int
										spacing: TAG_Int
								stronghold: TAG_Compound
									count: TAG_Int
									distance: TAG_Int
									spread: TAG_Int
							layers: TAG_List(TAG_Compound)
							lakes: TAG_Byte
							features: TAG_Byte
							biome: TAG_String
						type: TAG_String
					type: TAG_String
			bonus_chest: TAG_Byte
			generate_features: TAG_Byte
			seed: TAG_Long
		"""

		# create the generation settings
		generator_tag = TAG_Compound()
		generator_tag.name = "WorldGenSettings"

		# add basic tags
		generator_tag.tags.extend([
			TAG_Byte(name="bonus_chest", value=0),
			TAG_Byte(name="generate_features", value=0),
			TAG_Long(name="seed", value=self.seed)
		])

		# create the nested compound tags needed
		dimension_tag = TAG_Compound()
		dimension_tag.name = "dimensions"

		overworld_tag = TAG_Compound()
		overworld_tag.name = "minecraft:overworld"

		overworld_gen_tag = TAG_Compound()
		overworld_gen_tag.name = "generator"

		settings_tag = TAG_Compound()
		settings_tag.name = "settings"

		structures_tag = TAG_Compound()
		structures_tag.name = "structures"

		stronghold_tag = TAG_Compound()
		stronghold_tag.name = "stronghold"

		stronghold_tag.tags.extend([
			TAG_Int(name="count", value=128),
			TAG_Int(name="distance", value=32),
			TAG_Int(name="spread", value=3)
		])

		nested_structures = TAG_Compound()
		nested_structures.name = "structures"

		village_tag = TAG_Compound()
		village_tag.name = "minecraft:village"

		village_tag.tags.extend([
			TAG_Int(name="salt", value=randint(1e7, 1e9)),
			TAG_Int(name="separation", value=8),
			TAG_Int(name="spacing", value=32)
		])

		nested_structures.tags.append(village_tag)

		structures_tag.tags.extend([
			stronghold_tag,
			nested_structures
		])

		# add default values
		settings_tag.tags.extend([
			TAG_String(name="biome", value="minecraft:plains"),
			TAG_Byte(name="features", value=0),
			TAG_Byte(name="lakes", value=0),
			TAG_List(name="layers", type=TAG_Compound),
			structures_tag
		])

		# append each layer
		overworld_gen_tag.tags.extend([
			settings_tag,
			TAG_String(name="type", value="minecraft:flat")
		])

		overworld_tag.tags.extend([
			overworld_gen_tag,
			TAG_String(name="type", value="minecraft:overworld")
		])

		dimension_tag.tags.append(overworld_tag)
		
		end_tag = TAG_Compound()
		end_tag.name = "minecraft:the_end"
		
		end_gen_tag = TAG_Compound()
		end_gen_tag.name = "generator"
		
		end_biome_source = TAG_Compound()
		end_biome_source.name = "biome_source"
		
		end_biome_source.tags.extend([
			TAG_String(name="type", value="minecraft:the_end"),
			TAG_Long(name="seed", value=self.seed)
		])
		
		end_gen_tag.tags.extend([
			end_biome_source,
			TAG_Long(name="seed", value=self.seed),
			TAG_String(name="settings", value="minecraft:end"),
			TAG_String(name="type", value="minecraft:noise")
		])
		
		end_tag.tags.extend([
			end_gen_tag,
			TAG_String(name="type", value="minecraft:the_end")
		])
		
		nether_tag = TAG_Compound()
		nether_tag.name = "minecraft:the_nether"
		
		nether_gen_tag = TAG_Compound()
		nether_gen_tag.name = "generator"
		
		nether_biome_source = TAG_Compound()
		nether_biome_source.name = "biome_source"
		
		nether_biome_source.tags.extend([
			TAG_String(name="type", value="minecraft:multi_noise"),
			TAG_Long(name="seed", value=self.seed),
			TAG_String(name="preset", value="minecraft:nether")
		])
		
		nether_gen_tag.tags.extend([
			nether_biome_source,
			TAG_Long(name="seed", value=self.seed),
			TAG_String(name="settings", value="minecraft:nether"),
			TAG_String(name="type", value="minecraft:noise")
		])
		
		nether_tag.tags.extend([
			nether_gen_tag,
			TAG_String(name="type", value="minecraft:the_nether")
		])
		
		dimension_tag.tags.extend([end_tag, nether_tag])
		generator_tag.tags.append(dimension_tag)

		return generator_tag




"""
Copyright (c) 2010-2013 Thomas Woolford and contributors

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

"""
Handle the NBT (Named Binary Tag) data format

For more information about the NBT format:
https://minecraft.gamepedia.com/NBT_format

source: https://github.com/twoolie/NBT
"""

from struct import Struct, error as StructError
from gzip import GzipFile
from collections import MutableMapping, MutableSequence, Sequence
import sys

_PY3 = sys.version_info >= (3,)
if _PY3:
    unicode = str
    basestring = str
else:
    range = xrange

TAG_END = 0
TAG_BYTE = 1
TAG_SHORT = 2
TAG_INT = 3
TAG_LONG = 4
TAG_FLOAT = 5
TAG_DOUBLE = 6
TAG_BYTE_ARRAY = 7
TAG_STRING = 8
TAG_LIST = 9
TAG_COMPOUND = 10
TAG_INT_ARRAY = 11
TAG_LONG_ARRAY = 12


class MalformedFileError(Exception):
    """Exception raised on parse error."""
    pass


class TAG(object):
    """TAG, a variable with an intrinsic name."""
    id = None

    def __init__(self, value=None, name=None):
        self.name = name
        self.value = value

    # Parsers and Generators
    def _parse_buffer(self, buffer):
        raise NotImplementedError(self.__class__.__name__)

    def _render_buffer(self, buffer):
        raise NotImplementedError(self.__class__.__name__)

    # Printing and Formatting of tree
    def tag_info(self):
        """Return Unicode string with class, name and unnested value."""
        return self.__class__.__name__ + (
            '(%r)' % self.name if self.name
            else "") + ": " + self.valuestr()

    def valuestr(self):
        """Return Unicode string of unnested value. For iterators, this
        returns a summary."""
        return unicode(self.value)

    def namestr(self):
        """Return Unicode string of tag name."""
        return unicode(self.name)

    def pretty_tree(self, indent=0):
        """Return formated Unicode string of self, where iterable items are
        recursively listed in detail."""
        return ("\t" * indent) + self.tag_info()

    # Python 2 compatibility; Python 3 uses __str__ instead.
    def __unicode__(self):
        """Return a unicode string with the result in human readable format.
        Unlike valuestr(), the result is recursive for iterators till at least
        one level deep."""
        return unicode(self.value)

    def __str__(self):
        """Return a string (ascii formated for Python 2, unicode for Python 3)
        with the result in human readable format. Unlike valuestr(), the result
         is recursive for iterators till at least one level deep."""
        return str(self.value)

    # Unlike regular iterators, __repr__() is not recursive.
    # Use pretty_tree for recursive results.
    # iterators should use __repr__ or tag_info for each item, like
    #  regular iterators
    def __repr__(self):
        """Return a string (ascii formated for Python 2, unicode for Python 3)
        describing the class, name and id for debugging purposes."""
        return "<%s(%r) at 0x%x>" % (
            self.__class__.__name__, self.name, id(self))


class _TAG_Numeric(TAG):
    """_TAG_Numeric, comparable to int with an intrinsic name"""

    def __init__(self, value=None, name=None, buffer=None):
        super(_TAG_Numeric, self).__init__(value, name)
        if buffer:
            self._parse_buffer(buffer)

    # Parsers and Generators
    def _parse_buffer(self, buffer):
        # Note: buffer.read() may raise an IOError, for example if buffer is a
        # corrupt gzip.GzipFile
        self.value = self.fmt.unpack(buffer.read(self.fmt.size))[0]

    def _render_buffer(self, buffer):
        buffer.write(self.fmt.pack(self.value))


class _TAG_End(TAG):
    id = TAG_END
    fmt = Struct(">b")

    def _parse_buffer(self, buffer):
        # Note: buffer.read() may raise an IOError, for example if buffer is a
        # corrupt gzip.GzipFile
        value = self.fmt.unpack(buffer.read(1))[0]
        if value != 0:
            raise ValueError(
                "A Tag End must be rendered as '0', not as '%d'." % value)

    def _render_buffer(self, buffer):
        buffer.write(b'\x00')


# == Value Tags ==#
class TAG_Byte(_TAG_Numeric):
    """Represent a single tag storing 1 byte."""
    id = TAG_BYTE
    fmt = Struct(">b")


class TAG_Short(_TAG_Numeric):
    """Represent a single tag storing 1 short."""
    id = TAG_SHORT
    fmt = Struct(">h")


class TAG_Int(_TAG_Numeric):
    """Represent a single tag storing 1 int."""
    id = TAG_INT
    fmt = Struct(">i")
    """Struct(">i"), 32-bits integer, big-endian"""


class TAG_Long(_TAG_Numeric):
    """Represent a single tag storing 1 long."""
    id = TAG_LONG
    fmt = Struct(">q")


class TAG_Float(_TAG_Numeric):
    """Represent a single tag storing 1 IEEE-754 floating point number."""
    id = TAG_FLOAT
    fmt = Struct(">f")


class TAG_Double(_TAG_Numeric):
    """Represent a single tag storing 1 IEEE-754 double precision floating
    point number."""
    id = TAG_DOUBLE
    fmt = Struct(">d")


class TAG_Byte_Array(TAG, MutableSequence):
    """
    TAG_Byte_Array, comparable to a collections.UserList with
    an intrinsic name whose values must be bytes
    """
    id = TAG_BYTE_ARRAY

    def __init__(self, name=None, buffer=None):
        # TODO: add a value parameter as well
        super(TAG_Byte_Array, self).__init__(name=name)
        if buffer:
            self._parse_buffer(buffer)

    # Parsers and Generators
    def _parse_buffer(self, buffer):
        length = TAG_Int(buffer=buffer)
        self.value = bytearray(buffer.read(length.value))

    def _render_buffer(self, buffer):
        length = TAG_Int(len(self.value))
        length._render_buffer(buffer)
        buffer.write(bytes(self.value))

    # Mixin methods
    def __len__(self):
        return len(self.value)

    def __iter__(self):
        return iter(self.value)

    def __contains__(self, item):
        return item in self.value

    def __getitem__(self, key):
        return self.value[key]

    def __setitem__(self, key, value):
        # TODO: check type of value
        self.value[key] = value

    def __delitem__(self, key):
        del (self.value[key])

    def insert(self, key, value):
        # TODO: check type of value, or is this done by self.value already?
        self.value.insert(key, value)

    # Printing and Formatting of tree
    def valuestr(self):
        return "[%i byte(s)]" % len(self.value)

    def __unicode__(self):
        return '[' + ",".join([unicode(x) for x in self.value]) + ']'

    def __str__(self):
        return '[' + ",".join([str(x) for x in self.value]) + ']'


class TAG_Int_Array(TAG, MutableSequence):
    """
    TAG_Int_Array, comparable to a collections.UserList with
    an intrinsic name whose values must be integers
    """
    id = TAG_INT_ARRAY

    def __init__(self, name=None, buffer=None):
        # TODO: add a value parameter as well
        super(TAG_Int_Array, self).__init__(name=name)
        if buffer:
            self._parse_buffer(buffer)

    def update_fmt(self, length):
        """ Adjust struct format description to length given """
        self.fmt = Struct(">" + str(length) + "i")

    # Parsers and Generators
    def _parse_buffer(self, buffer):
        length = TAG_Int(buffer=buffer).value
        self.update_fmt(length)
        self.value = list(self.fmt.unpack(buffer.read(self.fmt.size)))

    def _render_buffer(self, buffer):
        length = len(self.value)
        self.update_fmt(length)
        TAG_Int(length)._render_buffer(buffer)
        buffer.write(self.fmt.pack(*self.value))

    # Mixin methods
    def __len__(self):
        return len(self.value)

    def __iter__(self):
        return iter(self.value)

    def __contains__(self, item):
        return item in self.value

    def __getitem__(self, key):
        return self.value[key]

    def __setitem__(self, key, value):
        self.value[key] = value

    def __delitem__(self, key):
        del (self.value[key])

    def insert(self, key, value):
        self.value.insert(key, value)

    # Printing and Formatting of tree
    def valuestr(self):
        return "[%i int(s)]" % len(self.value)


class TAG_Long_Array(TAG, MutableSequence):
    """
    TAG_Long_Array, comparable to a collections.UserList with
    an intrinsic name whose values must be integers
    """
    id = TAG_LONG_ARRAY

    def __init__(self, name=None, buffer=None):
        super(TAG_Long_Array, self).__init__(name=name)
        if buffer:
            self._parse_buffer(buffer)

    def update_fmt(self, length):
        """ Adjust struct format description to length given """
        self.fmt = Struct(">" + str(length) + "q")

    # Parsers and Generators
    def _parse_buffer(self, buffer):
        length = TAG_Int(buffer=buffer).value
        self.update_fmt(length)
        self.value = list(self.fmt.unpack(buffer.read(self.fmt.size)))

    def _render_buffer(self, buffer):
        length = len(self.value)
        self.update_fmt(length)
        TAG_Int(length)._render_buffer(buffer)
        buffer.write(self.fmt.pack(*self.value))

    # Mixin methods
    def __len__(self):
        return len(self.value)

    def __iter__(self):
        return iter(self.value)

    def __contains__(self, item):
        return item in self.value

    def __getitem__(self, key):
        return self.value[key]

    def __setitem__(self, key, value):
        self.value[key] = value

    def __delitem__(self, key):
        del (self.value[key])

    def insert(self, key, value):
        self.value.insert(key, value)

    # Printing and Formatting of tree
    def valuestr(self):
        return "[%i long(s)]" % len(self.value)


class TAG_String(TAG, Sequence):
    """
    TAG_String, comparable to a collections.UserString with an
    intrinsic name
    """
    id = TAG_STRING

    def __init__(self, value=None, name=None, buffer=None):
        super(TAG_String, self).__init__(value, name)
        if buffer:
            self._parse_buffer(buffer)

    # Parsers and Generators
    def _parse_buffer(self, buffer):
        length = TAG_Short(buffer=buffer)
        read = buffer.read(length.value)
        if len(read) != length.value:
            raise StructError()
        self.value = read.decode("utf-8")

    def _render_buffer(self, buffer):
        save_val = self.value.encode("utf-8")
        length = TAG_Short(len(save_val))
        length._render_buffer(buffer)
        buffer.write(save_val)

    # Mixin methods
    def __len__(self):
        return len(self.value)

    def __iter__(self):
        return iter(self.value)

    def __contains__(self, item):
        return item in self.value

    def __getitem__(self, key):
        return self.value[key]

    # Printing and Formatting of tree
    def __repr__(self):
        return self.value


# == Collection Tags ==#
class TAG_List(TAG, MutableSequence):
    """
    TAG_List, comparable to a collections.UserList with an intrinsic name
    """
    id = TAG_LIST

    def __init__(self, type=None, value=None, name=None, buffer=None):
        super(TAG_List, self).__init__(value, name)
        if type:
            self.tagID = type.id
        else:
            self.tagID = None
        self.tags = []
        if buffer:
            self._parse_buffer(buffer)
        # if self.tagID == None:
        #     raise ValueError("No type specified for list: %s" % (name))

    # Parsers and Generators
    def _parse_buffer(self, buffer):
        self.tagID = TAG_Byte(buffer=buffer).value
        self.tags = []
        length = TAG_Int(buffer=buffer)
        for x in range(length.value):
            self.tags.append(TAGLIST[self.tagID](buffer=buffer))

    def _render_buffer(self, buffer):
        TAG_Byte(self.tagID)._render_buffer(buffer)
        length = TAG_Int(len(self.tags))
        length._render_buffer(buffer)
        for i, tag in enumerate(self.tags):
            if tag.id != self.tagID:
                raise ValueError(
                    "List element %d(%s) has type %d != container type %d" %
                    (i, tag, tag.id, self.tagID))
            tag._render_buffer(buffer)

    # Mixin methods
    def __len__(self):
        return len(self.tags)

    def __iter__(self):
        return iter(self.tags)

    def __contains__(self, item):
        return item in self.tags

    def __getitem__(self, key):
        return self.tags[key]

    def __setitem__(self, key, value):
        self.tags[key] = value

    def __delitem__(self, key):
        del (self.tags[key])

    def insert(self, key, value):
        self.tags.insert(key, value)

    # Printing and Formatting of tree
    def __repr__(self):
        return "%i entries of type %s" % (
            len(self.tags), TAGLIST[self.tagID].__name__)

    # Printing and Formatting of tree
    def valuestr(self):
        return "[%i %s(s)]" % (len(self.tags), TAGLIST[self.tagID].__name__)

    def __unicode__(self):
        return "[" + ", ".join([tag.tag_info() for tag in self.tags]) + "]"

    def __str__(self):
        return "[" + ", ".join([tag.tag_info() for tag in self.tags]) + "]"

    def pretty_tree(self, indent=0):
        output = [super(TAG_List, self).pretty_tree(indent)]
        if len(self.tags):
            output.append(("\t" * indent) + "{")
            output.extend([tag.pretty_tree(indent + 1) for tag in self.tags])
            output.append(("\t" * indent) + "}")
        return '\n'.join(output)


class TAG_Compound(TAG, MutableMapping):
    """
    TAG_Compound, comparable to a collections.OrderedDict with an
    intrinsic name
    """
    id = TAG_COMPOUND

    def __init__(self, buffer=None, name=None):
        # TODO: add a value parameter as well
        super(TAG_Compound, self).__init__()
        self.tags = []
        if name:
            self.name = name
        else:
            self.name = ""
        if buffer:
            self._parse_buffer(buffer)

    # Parsers and Generators
    def _parse_buffer(self, buffer):
        while True:
            type = TAG_Byte(buffer=buffer)
            if type.value == TAG_END:
                # print("found tag_end")
                break
            else:
                name = TAG_String(buffer=buffer).value
                try:
                    tag = TAGLIST[type.value]()
                except KeyError:
                    raise ValueError("Unrecognised tag type %d" % type.value)
                tag.name = name
                self.tags.append(tag)
                tag._parse_buffer(buffer)

    def _render_buffer(self, buffer):
        for tag in self.tags:
            TAG_Byte(tag.id)._render_buffer(buffer)
            TAG_String(tag.name)._render_buffer(buffer)
            tag._render_buffer(buffer)
        buffer.write(b'\x00')  # write TAG_END

    # Mixin methods
    def __len__(self):
        return len(self.tags)

    def __iter__(self):
        for key in self.tags:
            yield key.name

    def __contains__(self, key):
        if isinstance(key, int):
            return key <= len(self.tags)
        elif isinstance(key, basestring):
            for tag in self.tags:
                if tag.name == key:
                    return True
            return False
        elif isinstance(key, TAG):
            return key in self.tags
        return False

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.tags[key]
        elif isinstance(key, basestring):
            for tag in self.tags:
                if tag.name == key:
                    return tag
            else:
                raise KeyError("Tag %s does not exist" % key)
        else:
            raise TypeError(
                "key needs to be either name of tag, or index of tag, "
                "not a %s" % type(key).__name__)

    def __setitem__(self, key, value):
        assert isinstance(value, TAG), "value must be an nbt.TAG"
        if isinstance(key, int):
            # Just try it. The proper error will be raised if it doesn't work.
            self.tags[key] = value
        elif isinstance(key, basestring):
            value.name = key
            for i, tag in enumerate(self.tags):
                if tag.name == key:
                    self.tags[i] = value
                    return
            self.tags.append(value)

    def __delitem__(self, key):
        if isinstance(key, int):
            del (self.tags[key])
        elif isinstance(key, basestring):
            self.tags.remove(self.__getitem__(key))
        else:
            raise ValueError(
                "key needs to be either name of tag, or index of tag")

    def keys(self):
        return [tag.name for tag in self.tags]

    def iteritems(self):
        for tag in self.tags:
            yield (tag.name, tag)

    # Printing and Formatting of tree
    def __unicode__(self):
        return "{" + ", ".join([tag.tag_info() for tag in self.tags]) + "}"

    def __str__(self):
        return "{" + ", ".join([tag.tag_info() for tag in self.tags]) + "}"

    def valuestr(self):
        return '{%i Entries}' % len(self.tags)

    def pretty_tree(self, indent=0):
        output = [super(TAG_Compound, self).pretty_tree(indent)]
        if len(self.tags):
            output.append(("\t" * indent) + "{")
            output.extend([tag.pretty_tree(indent + 1) for tag in self.tags])
            output.append(("\t" * indent) + "}")
        return '\n'.join(output)


TAGLIST = {TAG_END: _TAG_End, TAG_BYTE: TAG_Byte, TAG_SHORT: TAG_Short,
           TAG_INT: TAG_Int, TAG_LONG: TAG_Long, TAG_FLOAT: TAG_Float,
           TAG_DOUBLE: TAG_Double, TAG_BYTE_ARRAY: TAG_Byte_Array,
           TAG_STRING: TAG_String, TAG_LIST: TAG_List,
           TAG_COMPOUND: TAG_Compound, TAG_INT_ARRAY: TAG_Int_Array,
           TAG_LONG_ARRAY: TAG_Long_Array}


class NBTFile(TAG_Compound):
    """Represent an NBT file object."""

    def __init__(self, filename=None, buffer=None, fileobj=None):
        """
        Create a new NBTFile object.
        Specify either a filename, file object or data buffer.
        If filename of file object is specified, data should be GZip-compressed.
        If a data buffer is specified, it is assumed to be uncompressed.

        If filename is specified, the file is closed after reading and writing.
        If file object is specified, the caller is responsible for closing the
        file.
        """
        super(NBTFile, self).__init__()
        self.filename = filename
        self.type = TAG_Byte(self.id)
        closefile = True
        # make a file object
        if filename:
            self.filename = filename
            self.file = GzipFile(filename, 'rb')
        elif buffer:
            if hasattr(buffer, 'name'):
                self.filename = buffer.name
            self.file = buffer
            closefile = False
        elif fileobj:
            if hasattr(fileobj, 'name'):
                self.filename = fileobj.name
            self.file = GzipFile(fileobj=fileobj)
        else:
            self.file = None
            closefile = False
        # parse the file given initially
        if self.file:
            self.parse_file()
            if closefile:
                # Note: GzipFile().close() does NOT close the fileobj,
                # So we are still responsible for closing that.
                try:
                    self.file.close()
                except (AttributeError, IOError):
                    pass
            self.file = None

    def parse_file(self, filename=None, buffer=None, fileobj=None):
        """Completely parse a file, extracting all tags."""
        if filename:
            self.file = GzipFile(filename, 'rb')
        elif buffer:
            if hasattr(buffer, 'name'):
                self.filename = buffer.name
            self.file = buffer
        elif fileobj:
            if hasattr(fileobj, 'name'):
                self.filename = fileobj.name
            self.file = GzipFile(fileobj=fileobj)
        if self.file:
            try:
                type = TAG_Byte(buffer=self.file)
                if type.value == self.id:
                    name = TAG_String(buffer=self.file).value
                    self._parse_buffer(self.file)
                    self.name = name
                    self.file.close()
                else:
                    raise MalformedFileError(
                        "First record is not a Compound Tag")
            except StructError as e:
                raise MalformedFileError(
                    "Partial File Parse: file possibly truncated.")
        else:
            raise ValueError(
                "NBTFile.parse_file(): Need to specify either a "
                "filename or a file object"
            )

    def write_file(self, filename=None, buffer=None, fileobj=None):
        """Write this NBT file to a file."""
        closefile = True
        if buffer:
            self.filename = None
            self.file = buffer
            closefile = False
        elif filename:
            self.filename = filename
            self.file = GzipFile(filename, "wb")
        elif fileobj:
            self.filename = None
            self.file = GzipFile(fileobj=fileobj, mode="wb")
        elif self.filename:
            self.file = GzipFile(self.filename, "wb")
        elif not self.file:
            raise ValueError(
                "NBTFile.write_file(): Need to specify either a "
                "filename or a file object"
            )
        # Render tree to file
        TAG_Byte(self.id)._render_buffer(self.file)
        TAG_String(self.name)._render_buffer(self.file)
        self._render_buffer(self.file)
        # make sure the file is complete
        try:
            self.file.flush()
        except (AttributeError, IOError):
            pass
        if closefile:
            try:
                self.file.close()
            except (AttributeError, IOError):
                pass

    def __repr__(self):
        """
        Return a string (ascii formated for Python 2, unicode
        for Python 3) describing the class, name and id for
        debugging purposes.
        """
        if self.filename:
            return "<%s(%r) with %s(%r) at 0x%x>" % (
                self.__class__.__name__, self.filename,
                TAG_Compound.__name__, self.name, id(self)
            )
        else:
            return "<%s with %s(%r) at 0x%x>" % (
                self.__class__.__name__, TAG_Compound.__name__,
                self.name, id(self)
            )



from .chunk import Chunk
from math import ceil
from .block import BlockList
import time
import zlib


def cmod(n, base):
	"""
	Compute the modulus of a number like in languages such as C.
	
	:param n: the number to modulus
	:param base: the divisor
	"""

	return n - int(n / base) * base


class Region:
	"""
	A region class which defines an area of 32x32 chunks, or
	512x256x512 blocks. This makes up a single .mca file.
	"""

	def __init__(self, x, z):
		"""
		Creates the region class.

		:param x: the x position of the region
		:param z: the z position of the region
		"""

		self.chunks = self.create_chunks()
		self.x = x
		self.z = z

	
	def create_chunks(self):
		"""
		Creates the chunks for the region, setting them all to 0.
		"""

		chunks = []

		for x in range(32):
			chunks.append([])

			for z in range(32):
				chunks[len(chunks) - 1].append(0)

		return chunks


	def set_block(self, x, y, z, block):
		"""
		Sets a single block. This is called by the world's set_block.

		:param x: the x position of the block to set
		:param y: the y position of the block to set
		:param z: the z position of the block to set
		:param block: the block to set, has name and block properties
		"""

		# find the position of the chunk within the region
		chunkX = z // 16
		chunkZ = x // 16

		# if the chunk doesn't already exist, create it
		if not self.chunks[chunkX][chunkZ]:
			self.chunks[chunkX][chunkZ] = Chunk(
				z // 16,
				x // 16
			)

		return self.chunks[chunkX][chunkZ].set_block(
			cmod(z, 16), y, cmod(x, 16), block
		)

	
	def get_block(self, x, y, z):
		"""
		Get a single block in the region. Called by the World class.

		:param x: the x position of the block to get
		:param y: the y position of the block to get
		:param z: the z position of the block to get
		"""

		# find the position of the chunk within the region
		chunkX = z // 16
		chunkZ = x // 16

		# check if the chunk exists
		if not self.chunks[chunkX][chunkZ]:
			return BlockList.AIR.get_block()

		return self.chunks[chunkX][chunkZ].get_block(
			cmod(z, 16), y, cmod(x, 16)
		)


	def save(self, filename):
		"""
		Save the region to a file. The filename takes the form r.x.z.mca.
		The file format is specified in the Minecraft Wiki.

		:param filename: the filename to save the region to.
		"""

		# first we have to get the compressed NBT data from the chunks
		data = self.create_chunks()

		for x in range(32):
			for z in range(32):
				if self.chunks[x][z]:
					data[x][z] = self.chunks[x][z].get_compressed_nbt()
		
		# create the headers (chunk locations and timestamps)
		location_header = b""
		timestamp_header = b""
		all_data = b""
		current_offset = 2

		for x in range(32):
			for z in range(32):
				# add the time in epoch seconds to the timestamp header
				timestamp_header += int(time.time()).to_bytes(4, "big")

				# if the chunk is not initialised
				if data[x][z] == 0:
					location_header += b"\x00\x00\x00\x00"
					continue

				# get the chunk size in 4 KB sectors
				chunk_size = ceil((len(data) + 5) / 4096)

				# add the chunk offset and chunk size to the location header
				location_header += current_offset.to_bytes(3, "big")
				location_header += chunk_size.to_bytes(1, "big")

				# add the chunk size to the offset
				current_offset += chunk_size

				# add the chunk data to the data string
				all_data += data[x][z]

		with open(filename, "wb") as f:
			f.write(location_header + timestamp_header + all_data)


	def load(self, file_path):
		"""
		Load the region from a .mca file.

		:param file_path: the path to the .mca file
		"""

		with open(file_path, "rb") as f:
			data = f.read()

		# load the headers
		location_header = data[0:4096]
		timestamp_header = data[4096:8192]

		# go through each chunk in the file
		for x in range(32):
			for z in range(32):
				offset = 4 * (x + 32 * z)

				# find the location and size of the chunk in the file
				location = int.from_bytes(
					location_header[(offset):(offset + 3)], "big"
				)
				
				size = int.from_bytes(
					location_header[offset + 3], "big"
				)

				# if the chunk hasn't been loaded
				if location == 0:
					continue

				# load the data
				chunk_data = data[
					(location * 4096):(location * 4096 + size * 4096)
				]

				byte_size = int.from_bytes(chunk_data[0:4], "big")
				compression_type = int.from_bytes(chunk_data[4], "big")

				# if the chunk is stored in a format other than zlib
				if not compression_type == 2:
					raise ValueError("Chunk compression type not supported")

				uncompressed_data = zlib.decompress(
					chunk_data[5:(byte_size + 5)]
				)

				self.chunks[x][z] = Chunk(self.x * 32 + x, self.z * 32 + z)
				self.chunks[x][z].load(uncompressed_data)




from .nbt import *
from .section import Section
from .block import BlockList
import zlib
import io


class Long_Array(TAG_Long_Array):
	def __init__(self, name=""):
		super().__init__(name)


	def update_fmt(self, length):
		self.fmt = Struct(">" + str(length) + "Q")


class Chunk:
	"""
	Defines a single chunk of 16x256x16.
	1024 of these chunks make up a region.
	"""

	def __init__(self, x, z):
		"""
		Creates the chunk.

		:param x: the x position of the chunk
		:param z: the z position of the chunk
		"""

		self.x = x
		self.z = z
		self.sections = {}
	
	
	def set_block(self, x, y, z, block):
		"""
		Set a single block within the chunk.

		:param x: the x position of the block to set
		:param y: the y position of the block to set
		:param z: the z position of the block to set
		:param block: the type of block to set
		"""

		# find the section that contains the block
		sectionY = y >> 4

		if not sectionY in self.sections:
			self.sections[sectionY] = Section(self.x, sectionY, self.z)

		return self.sections[sectionY].set_block(x, y % 16, z, block)


	def get_block(self, x, y, z):
		"""
		Get a single block within the chunk.

		:param x: the x position of the block to get
		:param y: the y position of the block to get
		:param z: the z position of the block to get
		"""

		# find the section that contains the block
		sectionY = y >> 4

		# if the section does not exist, return air
		if not sectionY in self.sections:
			return BlockList.AIR

		return self.sections[sectionY].get_block(x, y % 16, z)


	def get_compressed_nbt(self):
		"""
		Takes the chunk data and puts it into an NBTFile object.
		Takes no parameters, returns NBTFile buffer.
		"""

		# create and name the file
		nbtfile = NBTFile()
		nbtfile.name = ""

		# create the root tag
		root_tag = TAG_Compound()
		root_tag.name = "Level"

		# create the necessary tags
		section_list = TAG_List(name="Sections", type=TAG_Compound)

		# add the sections
		for y in self.sections:
			section_list.tags.append(self.get_section_nbt(y))

		root_tag.tags.append(section_list)
		root_tag.tags.append(TAG_Int(name="xPos", value=self.x))
		root_tag.tags.append(TAG_Int(name="zPos", value=self.z))
		root_tag.tags.append(TAG_String(name="Status", value="full"))
		root_tag.tags.append(TAG_Byte(name="TerrainPopulated", value=1))
		root_tag.tags.append(TAG_Byte(name="LightPopulated", value=1))
		root_tag.tags.append(TAG_Byte(name="V", value=1))

		# write the root tag to the NBT file object
		nbtfile.tags.append(root_tag)

		# data version for 1.16.5
		nbtfile.tags.append(TAG_Int(name="DataVersion", value=2586))

		# create buffer
		bytes_buffer = io.BytesIO()

		# write to the buffer using NBT library
		nbtfile.write_file(buffer=bytes_buffer)

		# convert BytesIO to string
		bytes_buffer.seek(0)
		byte_str = bytes_buffer.read()

		# return the compressed data
		return self.pad_chunk_data(zlib.compress(byte_str))


	def get_section_nbt(self, y):
		"""
		Gets the NBT data for one section of the chunk.

		:param y: the y position of the section
		"""

		palette_tag = TAG_List(name="Palette", type=TAG_Compound)

		# create the blocks in the palette
		for block in self.sections[y].palette.blocks:
			block_tag = TAG_Compound()
			block_tag.tags.append(
				TAG_String(name="Name", value=block["name"])
			)

			# add each block's properties, if they exist
			if len(block["properties"]) > 0:
				properties_tag = TAG_Compound()
				properties_tag.name = "Properties"

				for prop in block["properties"]:
					properties_tag.tags.append(
						TAG_String(name=prop, value=block["properties"][prop])
					)

				block_tag.tags.append(properties_tag)

			# add the block to the palette
			palette_tag.tags.append(block_tag)

		# get an array of longs as block data
		block_data = self.sections[y].get_block_data()
		block_data_tag = Long_Array(name="BlockStates")
		block_data_tag.value = block_data

		section_tag = TAG_Compound()

		section_tag.tags.append(palette_tag)
		section_tag.tags.append(block_data_tag)
		section_tag.tags.append(TAG_Byte(name="Y", value=y))
		
		return section_tag


	def pad_chunk_data(self, data):
		"""
		Adds 5 bytes at the start describing the chunk, and pads the data
		with 0s so the length is a multiple of 4096.

		:param data: the binary string of compressed data
		"""

		data_length = len(data)
		new_length = data_length + 5
		padding = b"\x00" * (4096 - (new_length % 4096))

		return data_length.to_bytes(4, "big") + b"\x02" + data + padding


	def load(self, data):
		"""
		Load the chunk from the given data.

		:param data: the chunk data
		"""

		pass




from .nbt import *

class Option:
	"""
	Defines an option for the level.dat.
	"""

	def __init__(self, name: str, option_type, value: str):
		"""
		Creates the option.

		:param name: the name of the option.
		:param option_type: the type of the option.
		:param value: the option's value.
		"""

		self.name = name
		self.option_type = option_type
		self.value = value

	def create_tag(self):
		"""
		Creates the NBT tag and returns it. Takes no parameters.
		"""

		# check if it is a special type (eg. list)
		if self.option_type == TAG_Compound or \
			self.option_type == TAG_List or \
			self.option_type == TAG_Int_Array or \
			self.option_type == TAG_Long_Array:
			return False

		# create the tag
		tag = self.option_type(name=self.name, value=self.value)

		return tag

	"""
	List of valid options.
	"""
	VALID_OPTIONS = [
		"GameType", # the game mode - 0 survival, 1 creative, etc.
		"LevelName", # the world name
		"RandomSeed", # the world seed
		"SpawnX", # default spawn X
		"SpawnY", # default spawn Y
		"SpawnZ", # default spawn Z
		"allowCommands" # 1 for commands allowed, 0 for not allowed
	]

	# defines the possible tags for the options
	TAG_END = 0
	TAG_BYTE = 1
	TAG_SHORT = 2
	TAG_INT = 3
	TAG_LONG = 4
	TAG_FLOAT = 5
	TAG_DOUBLE = 6
	TAG_BYTE_ARRAY = 7
	TAG_STRING = 8
	TAG_LIST = 9
	TAG_COMPOUND = 10
	TAG_INT_ARRAY = 11
	TAG_LONG_ARRAY = 12




class GameRules:
	"""
	A list of gamerules used by the world.
	Contains all gamerules for Java edition as of 1.16.5.
	"""

	def __init__(self):
		"""
		Creates the rules of the class.
		"""

		self.rules = []
		self.create_game_rules()

	
	def create_game_rules(self):
		"""
		Saves every gamerule to the self.rules array.
		"""

		self.rules.extend([
			{
				"name": "announceAdvancements",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "commandBlockOutput",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "disableElytraMovementCheck",
				"value": "false",
				"type": "bool"
			},
			{
				"name": "disableRaids",
				"value": "false",
				"type": "bool"
			},
			{
				"name": "doDaylightCycle",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "doEntityDrops",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "doFireTick",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "doInsomnia",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "doImmediateRespawn",
				"value": "false",
				"type": "bool"
			},
			{
				"name": "doLimitedCrafting",
				"value": "false",
				"type": "bool"
			},
			{
				"name": "doMobLoot",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "doMobSpawning",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "doPatrolSpawning",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "doTileDrops",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "doTraderSpawning",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "doWeatherCycle",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "drowningDamage",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "fallDamage",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "fireDamage",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "forgiveDeadPlayers",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "freezeDamage",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "keepInventory",
				"value": "false",
				"type": "bool"
			},
			{
				"name": "logAdminCommands",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "maxCommandChainLength",
				"value": "65536",
				"type": "int"
			},
			{
				"name": "maxEntityCramming",
				"value": "24",
				"type": "int"
			},
			{
				"name": "mobGriefing",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "naturalRegeneration",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "playersSleepingPercentage",
				"value": "100",
				"type": "int"
			},
			{
				"name": "randomTickSpeed",
				"value": "3",
				"type": "int"
			},
			{
				"name": "reducedDebugInfo",
				"value": "false",
				"type": "bool"
			},
			{
				"name": "sendCommandFeedback",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "showDeathMessages",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "spawnRadius",
				"value": "10",
				"type": "int"
			},
			{
				"name": "spectatorsGenerateChunks",
				"value": "true",
				"type": "bool"
			},
			{
				"name": "universalAnger",
				"value": "false",
				"type": "bool"
			}
		])

	
	def set_rule(self, name, value):
		"""
		Sets a rule. The value must be the type (bool, int) of the gamerule.

		:param name: the name of the rule
		:param value: the value to set the rule to
		"""

		for rule in self.rules:
			if rule["name"] == name:
				if rule["type"] == "bool" and not isinstance(value, bool):
					raise ValueError("Given value is not of correct type")
				elif not isinstance(value, int):
					raise ValueError("Given value is not of correct type")

				rule["value"] = str(value).lower()
				return True

		raise ValueError(f"Game rule {name} could not be found")




import math


class ShapeList:
	"""
	A list of shapes used by the world.
	"""

	def __init__(self, world):
		"""
		Creates the list.

		:param world: the world to write to
		"""

		self.world = world
	
	
	def sphere(self, x, y, z, radius, block, full=False):
		"""
		Creates a sphere.

		:param x: the center x position of the sphere
		:param y: the center y position of the sphere
		:param z: the center z position of the sphere
		:param radius: the radius of the sphere
		:param block: the type of block the sphere is made of
		:param full: if the sphere should be filled
		"""

		# make sure the min/max y positions are in the world
		if y - radius < 0 or y + radius > 255:
			raise ValueError("Shape position is out of world boundaries")

		# cycle through all possible positions
		for sx in range(-radius, radius + 1):
			for sy in range(-radius, radius + 1):
				for sz in range(-radius, radius + 1):
					dist = math.sqrt(
						(sx + 0.5) * (sx + 0.5) +
						(sy + 0.5) * (sy + 0.5) +
						(sz + 0.5) * (sz + 0.5)
					)

					# if the distance from the center is the radius
					side_width = 0 if full else radius - 1

					if dist >= side_width and dist <= radius:
						self.world.set_block(
							x + sx,
							y + sy,
							z + sz,
							block
						)


	def cube(self, x, y, z, width, height, length, block):
		"""
		Creates a cube.

		:param x: the bottom left corner x position of the cube
		:param y: the bottom left corner y position of the cube
		:param z: the bottom left corner z position of the cube
		:param width: the width of the cube
		:param height: the height of the cube
		:param length: the length of the cube
		:param block: the material the cube should be
		"""

		for cx in range(width):
			for cy in range(height):
				for cz in range(length):
					self.world.set_block(
						x + cx,
						y + cy,
						z + cz,
						block
					)




from .palette import Palette
from .block import Air


class Section:
	"""
	Defines a single section of 16x16x16 blocks. There are 16 sections
	in a chunk.
	"""
	
	def __init__(self, x, y, z):
		"""
		Creates the section and the variables it needs.

		:param x: the x position of the section's chunk
		:param y: the y position of the section, from 0 to 16
		:param z: the z position of the section's chunk
		"""

		self.x = x
		self.y = y
		self.z = z

		self.palette = Palette()
		
		# add air to the palette
		self.palette.blocks.append(Air().get_block())
		self.palette.block_hashes.append("{}minecraft:air")
		self.palette.block_amounts.append(4096)

		self.blocks = [0] * 4096

	
	def set_block(self, x, y, z, block):
		"""
		Set a single block to the section. Adds the block to the
		palette if it is not already in there.

		:param x: the x position of the block to set
		:param y: the y position of the block to set
		:param z: the z position of the block to set
		:param block: the block type to set
		"""

		# remove the old block from the palette if it was the only instance
		is_sub = self.palette.subtract_block(self.palette.blocks[self.blocks[
			self.get_block_offset(x, y, z)
		]])

		if is_sub is not None:
			for i in range(len(self.blocks)):
				if self.blocks[i] > is_sub:
					self.blocks[i] -= 1

		# add the block to the palette
		block_index = self.palette.block_index(block.get_block())

		# update the section's block array
		self.blocks[self.get_block_offset(
			x, y, z
		)] = block_index

		return True


	def get_block(self, x, y, z):
		"""
		Get a single block within the section. Returns a dictionary with
		the name and properties of the block.

		:param x: the x position of the block to set
		:param y: the y position of the block to set
		:param z: the z position of the block to set
		"""

		# return the block within the palette
		block = self.palette.blocks[self.blocks[
			self.get_block_offset(x, y, z)
		]]

		return {
			"name": block["name"],
			"properties": block["properties"]
		}


	def get_block_offset(self, x, y, z):
		"""
		Get the offset of a certain block in the self.blocks array.

		:param x: the x position of the block relative to the section x
		:param y: the y position of the block relative to the section y
		:param z: the z position of the block relative to the section z
		"""

		return (x * (16 ** 0)) + (y * (16 ** 1)) + (z * (16 ** 2))


	def get_block_data(self):
		"""
		Create the blockdata for saving chunks. Each block takes up a minimum
		of 4 bits, but takes as many bits as needed to represent the biggest
		block index value in the palette. Takes no parameters, returns an
		array of longs (8 bytes each).
		"""

		long_array = [0]

		# the number of bits needed for each block
		bits_per_block = max(len(self.palette.blocks).bit_length(), 4)

		# the number of bits remaining in the current long
		bits_remaining = 64
		current_bit = 0

		# set the bits for each block
		# according to Minecraft Wiki, blocks are arranged in YZX format
		for y in range(16):
			for x in range(16):
				for z in range(16):
					# check if there are not enough bits in the long
					if bits_remaining < bits_per_block:
						long_array.append(0)
						bits_remaining = 64
						current_bit = 0

					# add the block value to the long
					long_array[-1] += self.blocks[
						self.get_block_offset(x, y, z)
					] << current_bit
					current_bit += bits_per_block
					bits_remaining -= bits_per_block

		return long_array
	



import json


class Palette:
	"""
	A palette of blocks for one section of a chunk. Each palette includes
	one instance of each unique block in the chunk.
	"""

	def __init__(self):
		"""
		Create an empty palette. Takes no arguments.
		"""

		self.blocks = []
		self.block_hashes = []
		self.block_amounts = []
	

	def block_index(self, block, add=True):
		"""
		Find the index of a block in the palette. Used for storing the blocks
		in the section.

		:param block: the block to search for
		:param add: whether to add the block if it is not found
		"""

		# create the unique hash of the block
		block_hash = self.generate_block_hash(block)

		# check if it is already in the palette
		if block_hash not in self.block_hashes and add:
			self.block_hashes.append(block_hash)
			self.blocks.append(block)
			self.block_amounts.append(1)
		elif block_hash not in self.block_hashes and not add:
			return -1
		elif block_hash in self.block_hashes and add:
			self.block_amounts[self.block_hashes.index(block_hash)] += 1

		return self.block_hashes.index(block_hash)


	def subtract_block(self, block):
		"""
		Remove a block from the palette if it no longer has any instances
		in the section. Returns the index if it is removed, None otherwise.

		:param block: the block to check
		"""

		block_hash = self.generate_block_hash(block)

		if block_hash not in self.block_hashes:
			return None

		index = self.block_hashes.index(block_hash)

		self.block_amounts[index] -= 1

		# if the block no longer has any instances
		if self.block_amounts[index] == 0:
			del self.block_amounts[index]
			del self.block_hashes[index]
			del self.blocks[index]
			return index

		return None


	def generate_block_hash(self, block):
		"""
		Generate a unique hash for each block. Includes the properties & name.

		:param block: the block to create the hash for
		"""

		return json.dumps(
			block["properties"],
			separators=(",", ":")
			) + block["name"]

